# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long, disable=protected-access
from collections import Counter, OrderedDict
from knack.log import get_logger
from azure.cli.core.util import CLIError
from azure.cli.core.commands import LongRunningOperation
from azure.cli.core.commands.client_factory import get_mgmt_service_client
from azure.cli.command_modules.network.zone_file.make_zone_file import make_zone_file
from azure.cli.command_modules.network.zone_file.parse_zone_file import parse_zone_file
from azure.cli.core.aaz import register_command
from azure.core.exceptions import HttpResponseError

from .aaz.latest.network.private_dns.link.vnet import Create as _PrivateDNSLinkVNetCreate
from .aaz.latest.network.private_dns.zone import Create as _PrivateDNSZoneCreate, Show as PrivateDNSZoneShow
from .aaz.latest.network.private_dns.record_set import Create as _RecordSetCreate, Delete as _RecordSetDelete, \
    ListByType as _RecordSetList, Show as _RecordSetShow, Update as _RecordSetUpdate

logger = get_logger(__name__)


# pylint: disable=too-many-statements, too-many-locals, too-many-branches
def import_zone(cmd, resource_group_name, private_zone_name, file_name):
    from azure.cli.core.util import read_file_content
    import sys
    from azure.mgmt.privatedns.models import RecordSet

    from azure.cli.core.azclierror import FileOperationError, UnclassifiedUserFault
    try:
        file_text = read_file_content(file_name)
    except FileNotFoundError:
        raise FileOperationError("No such file: " + str(file_name))
    except IsADirectoryError:
        raise FileOperationError("Is a directory: " + str(file_name))
    except PermissionError:
        raise FileOperationError("Permission denied: " + str(file_name))
    except OSError as e:
        raise UnclassifiedUserFault(e)

    zone_obj = parse_zone_file(file_text, private_zone_name)
    origin = private_zone_name
    record_sets = {}

    for record_set_name in zone_obj:
        for record_set_type in zone_obj[record_set_name]:
            record_set_obj = zone_obj[record_set_name][record_set_type]

            if record_set_type == 'soa':
                origin = record_set_name.rstrip('.')

            if not isinstance(record_set_obj, list):
                record_set_obj = [record_set_obj]

            for entry in record_set_obj:

                record_set_ttl = entry['ttl']
                record_set_key = '{}{}'.format(record_set_name.lower(), record_set_type)

                record = _build_record(cmd, entry)
                if not record:
                    logger.warning('Cannot import %s. RecordType is not found. Skipping...', entry['delim'].lower())
                    continue

                record_set = record_sets.get(record_set_key, None)
                if not record_set:

                    # Workaround for issue #2824
                    relative_record_set_name = record_set_name.rstrip('.')
                    if not relative_record_set_name.endswith(origin):
                        logger.warning(
                            'Cannot import %s. Only records relative to origin may be '
                            'imported at this time. Skipping...', relative_record_set_name)
                        continue

                    record_set = RecordSet(ttl=record_set_ttl)
                    record_sets[record_set_key] = record_set
                _add_record(record_set, record, record_set_type, is_list=record_set_type.lower() not in ['soa', 'cname'])

    total_records = 0
    for key, rs in record_sets.items():
        rs_name, rs_type = key.lower().rsplit('.', 1)
        rs_name = rs_name[:-(len(origin) + 1)] if rs_name != origin else '@'
        try:
            record_count = len(getattr(rs, _type_to_property_name(rs_type)))
        except TypeError:
            record_count = 1
        total_records += record_count
    cum_records = 0

    from azure.mgmt.privatedns import PrivateDnsManagementClient
    client = get_mgmt_service_client(cmd.cli_ctx, PrivateDnsManagementClient)

    print('== BEGINNING ZONE IMPORT: {} ==\n'.format(private_zone_name), file=sys.stderr)

    try:
        PrivateDNSZoneShow(cli_ctx=cmd.cli_ctx)(command_args={
            "name": private_zone_name,
            "resource_group": resource_group_name
        })

        logger.warning("Zone %s already exists in resource group %s.", private_zone_name, resource_group_name)
    except HttpResponseError:
        poller = PrivateDNSZoneCreate(cli_ctx=cmd.cli_ctx)(command_args={
            "name": private_zone_name,
            "resource_group": resource_group_name
        })
        result = LongRunningOperation(cmd.cli_ctx)(poller)
        if result["provisioningState"] != 'Succeeded':
            raise CLIError('Error occured while creating or updating private dns zone.')

    for key, rs in record_sets.items():

        rs_name, rs_type = key.lower().rsplit('.', 1)
        rs_name = '@' if rs_name == origin else rs_name
        if rs_name.endswith(origin):
            rs_name = rs_name[:-(len(origin) + 1)]

        try:
            record_count = len(getattr(rs, _type_to_property_name(rs_type)))
        except TypeError:
            record_count = 1
        if rs_name == '@' and rs_type == 'soa':
            root_soa = client.record_sets.get(resource_group_name, private_zone_name, 'soa', '@')
            rs.soa_record.host = root_soa.soa_record.host
            rs_name = '@'
        try:
            client.record_sets.create_or_update(
                resource_group_name, private_zone_name, rs_type, rs_name, rs)
            cum_records += record_count
            print("({}/{}) Imported {} records of type '{}' and name '{}'"
                  .format(cum_records, total_records, record_count, rs_type, rs_name), file=sys.stderr)
        except HttpResponseError as ex:
            logger.error(ex)
    print("\n== {}/{} RECORDS IMPORTED SUCCESSFULLY: '{}' =="
          .format(cum_records, total_records, private_zone_name), file=sys.stderr)


# pylint: disable=too-many-branches
def export_zone(cmd, resource_group_name, private_zone_name, file_name=None):
    from azure.mgmt.privatedns import PrivateDnsManagementClient
    from time import localtime, strftime
    client = get_mgmt_service_client(cmd.cli_ctx, PrivateDnsManagementClient).record_sets
    record_sets_soa = client.list_by_type(resource_group_name, private_zone_name, "soa")
    record_sets_all = client.list(resource_group_name, private_zone_name)
    zone_obj = OrderedDict({
        '$origin': private_zone_name.rstrip('.') + '.',
        'resource-group': resource_group_name,
        'zone-name': private_zone_name.rstrip('.'),
        'datetime': strftime('%a, %d %b %Y %X %z', localtime())
    })

    for record_set in record_sets_soa:
        record_type = record_set.type.rsplit('/', 1)[1].lower()
        if record_type == 'soa':
            record_set_name = record_set.name
            record_data = getattr(record_set, _type_to_property_name(record_type), None)

            if not isinstance(record_data, list):
                record_data = [record_data]

            for record in record_data:

                record_obj = {'ttl': record_set.ttl}

                if record_type == 'soa':
                    zone_obj[record_set_name] = OrderedDict()
                    zone_obj[record_set_name][record_type] = []
                    record_obj.update({
                        'mname': record.host.rstrip('.') + '.',
                        'rname': record.email.rstrip('.') + '.',
                        'serial': int(record.serial_number), 'refresh': record.refresh_time,
                        'retry': record.retry_time, 'expire': record.expire_time,
                        'minimum': record.minimum_ttl
                    })
                    zone_obj['$ttl'] = record.minimum_ttl
                    zone_obj[record_set_name][record_type].append(record_obj)

    for record_set in record_sets_all:
        record_type = record_set.type.rsplit('/', 1)[1].lower()
        record_set_name = record_set.name
        record_data = getattr(record_set, _type_to_property_name(record_type), None)

        # ignore empty record sets
        if not record_data:
            continue

        if not isinstance(record_data, list):
            record_data = [record_data]

        if record_set_name not in zone_obj:
            zone_obj[record_set_name] = OrderedDict()

        for record in record_data:

            record_obj = {'ttl': record_set.ttl}

            if record_type not in zone_obj[record_set_name]:
                zone_obj[record_set_name][record_type] = []

            if record_type == 'aaaa':
                record_obj.update({'ip': record.ipv6_address})
            elif record_type == 'a':
                record_obj.update({'ip': record.ipv4_address})
            elif record_type == 'caa':
                record_obj.update({'val': record.value, 'tag': record.tag, 'flags': record.flags})
            elif record_type == 'cname':
                record_obj.update({'alias': record.cname.rstrip('.') + '.'})
            elif record_type == 'mx':
                record_obj.update({'preference': record.preference, 'host': record.exchange})
            elif record_type == 'ns':
                record_obj.update({'host': record.nsdname})
            elif record_type == 'ptr':
                record_obj.update({'host': record.ptrdname})
            elif record_type == 'soa':
                continue
            elif record_type == 'srv':
                record_obj.update({'priority': record.priority, 'weight': record.weight,
                                   'port': record.port, 'target': record.target})
            elif record_type == 'txt':
                record_obj.update({'txt': ''.join(record.value)})

            zone_obj[record_set_name][record_type].append(record_obj)

    zone_file_content = make_zone_file(zone_obj)
    print(zone_file_content)
    if file_name:
        try:
            with open(file_name, 'w') as f:
                f.write(zone_file_content)
        except OSError:
            raise CLIError('Unable to export to file: {}'.format(file_name))


# pylint: disable=too-many-return-statements, inconsistent-return-statements, unused-argument
def _build_record(cmd, data):
    from azure.mgmt.privatedns.models import AaaaRecord, ARecord, CnameRecord, MxRecord, PtrRecord, SoaRecord, SrvRecord, TxtRecord
    record_type = data['delim'].lower()
    try:
        if record_type == 'aaaa':
            return AaaaRecord(ipv6_address=data['ip'])
        if record_type == 'a':
            return ARecord(ipv4_address=data['ip'])
        if record_type == 'cname':
            return CnameRecord(cname=data['alias'])
        if record_type == 'mx':
            return MxRecord(preference=data['preference'], exchange=data['host'])
        if record_type == 'ptr':
            return PtrRecord(ptrdname=data['host'])
        if record_type == 'soa':
            return SoaRecord(host=data['host'], email=data['email'], serial_number=data['serial'],
                             refresh_time=data['refresh'], retry_time=data['retry'], expire_time=data['expire'],
                             minimum_ttl=data['minimum'])
        if record_type == 'srv':
            return SrvRecord(
                priority=int(data['priority']), weight=int(data['weight']), port=int(data['port']),
                target=data['target'])
        if record_type in ['txt', 'spf']:
            text_data = data['txt']
            return TxtRecord(value=text_data) if isinstance(text_data, list) else TxtRecord(value=[text_data])
    except KeyError as ke:
        raise CLIError("The {} record '{}' is missing a property.  {}"
                       .format(record_type, data['name'], ke))


class PrivateDNSZoneCreate(_PrivateDNSZoneCreate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.if_none_match._registered = False
        args_schema.location._registered = False

        return args_schema

    def pre_operations(self):
        args = self.ctx.args
        if args.name.to_serialized_data().endswith(".local"):
            logger.warning(
                "Please be aware that DNS names ending with `.local` are reserved for use with multicast DNS and "
                "may not work as expected with some operating systems. "
                "For details refer to your operating systems documentation."
            )
        args.location = "global"
        args.if_none_match = "*"


class PrivateDNSLinkVNetCreate(_PrivateDNSLinkVNetCreate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        from azure.cli.core.aaz import AAZResourceIdArgFormat
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.virtual_network._fmt = AAZResourceIdArgFormat(
            template="/subscriptions/{subscription}/resourceGroups/{resource_group}/providers/Microsoft.Network"
                     "/virtualNetworks/{}"
        )
        args_schema.registration_enabled._required = True
        args_schema.virtual_network._required = True
        args_schema.if_none_match._registered = False
        args_schema.location._registered = False

        return args_schema

    def pre_operations(self):
        args = self.ctx.args
        args.location = "global"
        args.if_none_match = "*"


# region RecordSetCreate
class RecordSetCreate(_RecordSetCreate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.record_type._required = False
        args_schema.record_type._registered = False
        args_schema.if_none_match._registered = False
        args_schema.a_records._registered = False
        args_schema.aaaa_records._registered = False
        args_schema.cname_record._registered = False
        args_schema.mx_records._registered = False
        args_schema.ptr_records._registered = False
        args_schema.soa_record._registered = False
        args_schema.srv_records._registered = False
        args_schema.txt_records._registered = False

        return args_schema


@register_command("network private-dns record-set a create")
class RecordSetACreate(RecordSetCreate):
    """ Create an empty A record set.

    :example: Create an empty A record set.
        az network private-dns record-set a create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "A"
        args.if_none_match = "*"


@register_command("network private-dns record-set aaaa create")
class RecordSetAAAACreate(RecordSetCreate):
    """ Create an empty AAAA record set.

    :example: Create an empty AAAA record set.
        az network private-dns record-set aaaa create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "AAAA"
        args.if_none_match = "*"


@register_command("network private-dns record-set cname create")
class RecordSetCNAMECreate(RecordSetCreate):
    """ Create an empty CNAME record set.

    :example: Create an empty CNAME record set.
        az network private-dns record-set cname create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "CNAME"
        args.if_none_match = "*"


@register_command("network private-dns record-set mx create")
class RecordSetMXCreate(RecordSetCreate):
    """ Create an empty MX record set.

    :example: Create an empty MX record set.
        az network private-dns record-set mx create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "MX"
        args.if_none_match = "*"


@register_command("network private-dns record-set ptr create")
class RecordSetPTRCreate(RecordSetCreate):
    """ Create an empty PTR record set.

    :example: Create an empty PTR record set.
        az network private-dns record-set ptr create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "PTR"
        args.if_none_match = "*"


@register_command("network private-dns record-set srv create")
class RecordSetSRVCreate(RecordSetCreate):
    """ Create an empty SRV record set.

    :example: Create an empty SRV record set.
        az network private-dns record-set srv create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SRV"
        args.if_none_match = "*"


@register_command("network private-dns record-set txt create")
class RecordSetTXTCreate(RecordSetCreate):
    """ Create an empty TXT record set.

    :example: Create an empty TXT record set.
        az network private-dns record-set txt create -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "TXT"
        args.if_none_match = "*"
# endregion RecordSetCreate


# region RecordSetDelete
class RecordSetDelete(_RecordSetDelete):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.record_type._required = False
        args_schema.record_type._registered = False

        return args_schema


@register_command("network private-dns record-set a delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetADelete(RecordSetDelete):
    """ Delete an A record set and all associated records.

    :example: Delete an A record set and all associated records.
        az network private-dns record-set a delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "A"


@register_command("network private-dns record-set aaaa delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetAAAADelete(RecordSetDelete):
    """ Delete an AAAA record set and all associated records.

    :example: Delete an AAAA record set and all associated records.
        az network private-dns record-set aaaa delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "AAAA"


@register_command("network private-dns record-set cname delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetCNAMEDelete(RecordSetDelete):
    """ Delete a CNAME record set and its associated record.

    :example: Delete a CNAME record set and its associated record.
        az network private-dns record-set cname delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "CNAME"


@register_command("network private-dns record-set mx delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetMXDelete(RecordSetDelete):
    """ Delete an MX record set and all associated records.

    :example: Delete an MX record set and all associated records.
        az network private-dns record-set mx delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "MX"


@register_command("network private-dns record-set ptr delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetPTRDelete(RecordSetDelete):
    """ Delete a PTR record set and all associated records.

    :example: Delete a PTR record set and all associated records.
        az network private-dns record-set ptr delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "PTR"


@register_command("network private-dns record-set srv delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetSRVDelete(RecordSetDelete):
    """ Delete an SRV record set and all associated records.

    :example: Delete an SRV record set and all associated records.
        az network private-dns record-set srv delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SRV"


@register_command("network private-dns record-set txt delete", confirmation="Are you sure you want to perform this operation?")
class RecordSetTXTDelete(RecordSetDelete):
    """ Delete a TXT record set and all associated records.

    :example: Delete a TXT record set and all associated records.
        az network private-dns record-set txt delete -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "TXT"
# endregion RecordSetDelete


# region RecordSetList
class RecordSetList(_RecordSetList):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.record_type._required = False
        args_schema.record_type._registered = False

        return args_schema


@register_command("network private-dns record-set a list")
class RecordSetAList(RecordSetList):
    """ List all A record sets in a zone.

    :example: List all A record sets in a zone.
        az network private-dns record-set a list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "A"


@register_command("network private-dns record-set aaaa list")
class RecordSetAAAAList(RecordSetList):
    """ List all AAAA record sets in a zone.

    :example: List all AAAA record sets in a zone.
        az network private-dns record-set aaaa list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "AAAA"


@register_command("network private-dns record-set cname list")
class RecordSetCNAMEList(RecordSetList):
    """ List the CNAME record set in a zone.

    :example: List the CNAME record set in a zone.
        az network private-dns record-set cname list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "CNAME"


@register_command("network private-dns record-set mx list")
class RecordSetMXList(RecordSetList):
    """ List all MX record sets in a zone.

    :example: List all MX record sets in a zone.
        az network private-dns record-set mx list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "MX"


@register_command("network private-dns record-set ptr list")
class RecordSetPTRList(RecordSetList):
    """ List all PTR record sets in a zone.

    :example: List all PTR record sets in a zone.
        az network private-dns record-set ptr list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "PTR"


@register_command("network private-dns record-set srv list")
class RecordSetSRVList(RecordSetList):
    """ List all SRV record sets in a zone.

    :example: List all SRV record sets in a zone.
        az network private-dns record-set srv list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SRV"


@register_command("network private-dns record-set txt list")
class RecordSetTXTList(RecordSetList):
    """ List all TXT record sets in a zone.

    :example: List all TXT record sets in a zone.
        az network private-dns record-set txt list -g MyResourceGroup -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "TXT"
# endregion RecordSetList


# region RecordSetShow
class RecordSetShow(_RecordSetShow):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.record_type._required = False
        args_schema.record_type._registered = False

        return args_schema


@register_command("network private-dns record-set a show")
class RecordSetAShow(RecordSetShow):
    """ Get the details of an A record set.

    :example: Get the details of an A record set.
        az network private-dns record-set a show -g MyResourceGroup -n MyRecordSet -z www.mysite.com
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "A"


@register_command("network private-dns record-set aaaa show")
class RecordSetAAAAShow(RecordSetShow):
    """ Get the details of an AAAA record set.

    :example: Get the details of an AAAA record set.
        az network private-dns record-set aaaa show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "AAAA"


@register_command("network private-dns record-set cname show")
class RecordSetCNAMEShow(RecordSetShow):
    """ Get the details of a CNAME record set.

    :example: Get the details of a CNAME record set.
        az network private-dns record-set cname show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "CNAME"


@register_command("network private-dns record-set mx show")
class RecordSetMXShow(RecordSetShow):
    """ Get the details of an MX record set.

    :example: Get the details of an MX record set.
        az network private-dns record-set mx show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "MX"


@register_command("network private-dns record-set ptr show")
class RecordSetPTRShow(RecordSetShow):
    """ Get the details of a PTR record set.

    :example: Get the details of a PTR record set.
        az network private-dns record-set ptr show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "PTR"


@register_command("network private-dns record-set soa show")
class RecordSetSOAShow(RecordSetShow):
    """ Get the details of an SOA record.

    :example: Get the details of an SOA record.
        az network private-dns record-set soa show -g MyResourceGroup -z www.mysite.com
    """
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.name._required = False
        args_schema.name._registered = False

        return args_schema

    def pre_operations(self):
        args = self.ctx.args
        args.name = "@"
        args.record_type = "SOA"


@register_command("network private-dns record-set srv show")
class RecordSetSRVShow(RecordSetShow):
    """ Get the details of an SRV record set.

    :example: Get the details of an SRV record set.
        az network private-dns record-set srv show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SRV"


@register_command("network private-dns record-set txt show")
class RecordSetTXTShow(RecordSetShow):
    """ Get the details of a TXT record set.

    :example: Get the details of a TXT record set.
        az network private-dns record-set txt show -g MyResourceGroup -z www.mysite.com -n MyRecordSet
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "TXT"
# endregion RecordSetShow


# region RecordSetUpdate
class RecordSetUpdate(_RecordSetUpdate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.record_type._required = False
        args_schema.record_type._registered = False
        args_schema.a_records._registered = False
        args_schema.aaaa_records._registered = False
        args_schema.cname_record._registered = False
        args_schema.mx_records._registered = False
        args_schema.ptr_records._registered = False
        args_schema.soa_record._registered = False
        args_schema.srv_records._registered = False
        args_schema.txt_records._registered = False

        return args_schema


@register_command("network private-dns record-set a update")
class RecordSetAUpdate(RecordSetUpdate):
    """ Update an A record set.

    :example: Update an A record set.
        az network private-dns record-set a update -g MyResourceGroup -n MyRecordSet -z www.mysite.com --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "A"


@register_command("network private-dns record-set aaaa update")
class RecordSetAAAAUpdate(RecordSetUpdate):
    """ Update an AAAA record set.

    :example: Update an AAAA record set.
        az network private-dns record-set aaaa update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "AAAA"


@register_command("network private-dns record-set cname update")
class RecordSetCNAMEUpdate(RecordSetUpdate):
    """ Update a CNAME record set.

    :example: Update a CNAME record set.
        az network private-dns record-set cname update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "CNAME"


@register_command("network private-dns record-set mx update")
class RecordSetMXUpdate(RecordSetUpdate):
    """ Update an MX record set.

    :example: Update an MX record set.
        az network private-dns record-set mx update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "MX"


@register_command("network private-dns record-set ptr update")
class RecordSetPTRUpdate(RecordSetUpdate):
    """ Update a PTR record set.

    :example: Update a PTR record set.
        az network private-dns record-set ptr update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "PTR"


class RecordSetSOAUpdate(RecordSetUpdate):
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SOA"


@register_command("network private-dns record-set srv update")
class RecordSetSRVUpdate(RecordSetUpdate):
    """ Update an SRV record set.

    :example: Update an SRV record set.
        az network private-dns record-set srv update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "SRV"


@register_command("network private-dns record-set txt update")
class RecordSetTXTUpdate(RecordSetUpdate):
    """ Update a TXT record set.

    :example: Update a TXT record set.
        az network private-dns record-set txt update -g MyResourceGroup -z www.mysite.com -n MyRecordSet --metadata owner=WebTeam
    """
    def pre_operations(self):
        args = self.ctx.args
        args.record_type = "TXT"
# endregion RecordSetUpdate


def _type_to_property_name(key):
    type_dict = {
        'a': 'a_records',
        'aaaa': 'aaaa_records',
        'cname': 'cname_record',
        'mx': 'mx_records',
        'ptr': 'ptr_records',
        'soa': 'soa_record',
        'srv': 'srv_records',
        'txt': 'txt_records',
    }
    return type_dict[key.lower()]


def _add_record(record_set, record, record_type, is_list=False):
    record_property = _type_to_property_name(record_type)

    if is_list:
        record_list = getattr(record_set, record_property)
        if record_list is None:
            setattr(record_set, record_property, [])
            record_list = getattr(record_set, record_property)
        record_list.append(record)
    else:
        setattr(record_set, record_property, record)


def _to_snake(s):
    import re
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)

    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def _convert_to_snake_case(element):
    if isinstance(element, dict):
        ret = {}
        for k, v in element.items():
            ret[_to_snake(k)] = _convert_to_snake_case(v)

        return ret

    if isinstance(element, list):
        return [_convert_to_snake_case(i) for i in element]

    return element


def _record_show_func(record_type):
    return globals()["RecordSet{}Show".format(record_type.upper())]


def _record_create_func(record_type):
    return globals()["RecordSet{}Create".format(record_type.upper())]


def _record_delete_func(record_type):
    return globals()["RecordSet{}Delete".format(record_type.upper())]


def _record_update_func(record_type):
    return globals()["RecordSet{}Update".format(record_type.upper())]


def _privatedns_type_to_property_name(key):
    type_dict = {
        # `record_type`: (`snake_case`, `camel_case`)
        'a': ('a_records', "aRecords"),
        'aaaa': ('aaaa_records', "aaaaRecords"),
        'cname': ('cname_record', "cnameRecord"),
        'mx': ('mx_records', "mxRecords"),
        'ptr': ('ptr_records', "ptrRecords"),
        'soa': ('soa_record', "soaRecord"),
        'srv': ('srv_records', "srvRecords"),
        'txt': ('txt_records', "txtRecords")
    }
    return type_dict[key.lower()]


def _privatedns_add_record(record_set, record, record_type, is_list=False):
    record_property, _ = _privatedns_type_to_property_name(record_type)

    if is_list:
        record_list = record_set.setdefault(record_property, [])
        record_list.append(record)
    else:
        record_set[record_property] = record


def _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, is_list=True):
    record_snake, record_camel = _privatedns_type_to_property_name(record_type)
    is_empty = False

    try:
        _record_show = _record_show_func(record_type)
        ret = _record_show(cli_ctx=cmd.cli_ctx)(command_args={
            "resource_group": resource_group_name,
            "zone_name": private_zone_name,
            "record_type": record_type,
            "name": relative_record_set_name
        })
        record_set = {}
        record_set["ttl"] = ret.get("ttl", None)
        record_set[record_snake] = ret.get(record_camel, None)
        record_set = _convert_to_snake_case(record_set)
    except HttpResponseError:
        record_set = {"ttl": 3600}
        is_empty = True

    _privatedns_add_record(record_set, record, record_type, is_list)

    if is_empty:
        _record_create = _record_create_func(record_type)
        return _record_create(cli_ctx=cmd.cli_ctx)(command_args={
            "resource_group": resource_group_name,
            "zone_name": private_zone_name,
            "record_type": record_type,
            "name": relative_record_set_name,
            **record_set
        })

    _record_update = _record_update_func(record_type)
    return _record_update(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "zone_name": private_zone_name,
        "record_type": record_type,
        "name": relative_record_set_name,
        **record_set
    })


def add_privatedns_aaaa_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, ipv6_address):
    record = {"ipv6_address": ipv6_address}
    record_type = 'aaaa'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_a_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, ipv4_address):
    record = {"ipv4_address": ipv4_address}
    record_type = 'a'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_mx_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, preference, exchange):
    record = {"preference": int(preference), "exchange": exchange}
    record_type = 'mx'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_ptr_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, dname):
    record = {"ptrdname": dname}
    record_type = 'ptr'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_srv_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, priority, weight, port, target):
    record = {"priority": priority, "weight": weight, "port": port, "target": target}
    record_type = 'srv'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_txt_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, value):
    record = {"value": value}
    record_type = 'txt'
    long_text = ''.join(x for x in record["value"])
    original_len = len(long_text)
    record["value"] = []
    while len(long_text) > 255:
        record["value"].append(long_text[:255])
        long_text = long_text[255:]
    record["value"].append(long_text)
    final_str = ''.join(record["value"])
    final_len = len(final_str)
    assert original_len == final_len
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name)


def add_privatedns_cname_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, cname):
    record = {"cname": cname}
    record_type = 'cname'
    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, is_list=False)


def update_privatedns_soa_record(cmd, resource_group_name, private_zone_name, host=None, email=None,
                                 serial_number=None, refresh_time=None, retry_time=None, expire_time=None,
                                 minimum_ttl=None):
    relative_record_set_name = '@'
    record_type = 'soa'

    record_set = RecordSetSOAShow(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "zone_name": private_zone_name,
        "record_type": record_type,
        "name": relative_record_set_name
    })

    record_camal = record_set["soaRecord"]
    record = {}
    record["host"] = host or record_camal.get("host", None)
    record["email"] = email or record_camal.get("email", None)
    record["serial_number"] = serial_number or record_camal.get("serialNumber", None)
    record["refresh_time"] = refresh_time or record_camal.get("refreshTime", None)
    record["retry_time"] = retry_time or record_camal.get("retryTime", None)
    record["expire_time"] = expire_time or record_camal.get("expireTime", None)
    record["minimum_ttl"] = minimum_ttl or record_camal.get("minimumTTL", None)

    return _privatedns_add_save_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, is_list=False)


def remove_privatedns_aaaa_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, ipv6_address, keep_empty_record_set=False):
    record = {"ipv6_address": ipv6_address}
    record_type = 'aaaa'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_a_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, ipv4_address, keep_empty_record_set=False):
    record = {"ipv4_address": ipv4_address}
    record_type = 'a'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_cname_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, cname, keep_empty_record_set=False):
    record = {"cname": cname}
    record_type = 'cname'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, is_list=False, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_mx_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, preference, exchange, keep_empty_record_set=False):
    record = {"preference": int(preference), "exchange": exchange}
    record_type = 'mx'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_ptr_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, dname, keep_empty_record_set=False):
    record = {"ptrdname": dname}
    record_type = 'ptr'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_srv_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, priority, weight, port, target, keep_empty_record_set=False):
    record = {"priority": priority, "weight": weight, "port": port, "target": target}
    record_type = 'srv'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def remove_privatedns_txt_record(cmd, resource_group_name, private_zone_name, relative_record_set_name, value, keep_empty_record_set=False):
    record = {"value": value}
    record_type = 'txt'
    return _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set=keep_empty_record_set)


def _privatedns_remove_record(cmd, record, record_type, relative_record_set_name, resource_group_name, private_zone_name, keep_empty_record_set, is_list=True):
    record_snake, record_camel = _privatedns_type_to_property_name(record_type)
    _record_show = _record_show_func(record_type)
    ret = _record_show(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "zone_name": private_zone_name,
        "record_type": record_type,
        "name": relative_record_set_name
    })
    record_set = {}
    record_set["ttl"] = ret.get("ttl", None)
    record_set[record_snake] = ret.get(record_camel, None)
    record_set = _convert_to_snake_case(record_set)

    if is_list:
        record_list = record_set[record_snake]
        if record_list is not None:
            keep_list = [r for r in record_list if not dict_matches_filter(r, record)]
            if len(keep_list) == len(record_list):
                raise CLIError('Record {} not found.'.format(str(record)))

            record_set[record_snake] = keep_list
    else:
        record_set[record_snake] = None

    if is_list:
        records_remaining = len(record_set[record_snake]) if record_set[record_snake] is not None else 0
    else:
        records_remaining = 1 if record_set[record_snake] is not None else 0

    if not records_remaining and not keep_empty_record_set:
        logger.info('Removing empty %s record set: %s', record_type, relative_record_set_name)

        _record_delete = _record_delete_func(record_type)
        return _record_delete(cli_ctx=cmd.cli_ctx)(command_args={
            "resource_group": resource_group_name,
            "zone_name": private_zone_name,
            "record_type": record_type,
            "name": relative_record_set_name
        })

    _record_update = _record_update_func(record_type)
    return _record_update(cli_ctx=cmd.cli_ctx)(command_args={
        "resource_group": resource_group_name,
        "zone_name": private_zone_name,
        "record_type": record_type,
        "name": relative_record_set_name,
        **record_set
    })


def dict_matches_filter(d, filter_dict):
    sentinel = object()
    return all(not filter_dict.get(key, None)
               or str(filter_dict[key]) == str(d.get(key, sentinel)) # noqa
               or lists_match(filter_dict[key], d.get(key, [])) for key in filter_dict)  # noqa


# pylint: disable=too-many-function-args
def lists_match(l1, l2):
    try:
        return Counter(l1) == Counter(l2)
    except TypeError:
        return False



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import uuid
from azure.mgmt.advisor.models import SuppressionContract


def list_recommendations(client, ids=None, resource_group_name=None,
                         category=None, refresh=None):
    if refresh:
        _generate_recommendations(client)
    scope = _build_filter_string(ids, resource_group_name, category)
    return client.list(scope)


def disable_recommendations(client, ids=None, recommendation_name=None,
                            resource_group_name=None, days=None):
    if recommendation_name is None and ids is None:
        from knack.prompting import prompt_y_n

        if not prompt_y_n("\nAre you sure you want to disabled all recommendations?"):
            return None

    recs = _get_recommendations(
        client=client.recommendations,
        ids=ids,
        resource_group_name=resource_group_name,
        recommendation_name=recommendation_name)

    if recs is not None:
        for rec in recs:
            suppression_name = str(uuid.uuid4())
            ttl = '{}:00:00:00'.format(days) if days else ''

            result = _parse_recommendation_uri(rec.id)
            resource_uri = result['resource_uri']
            recommendation_id = result['recommendation_id']
            suppression_contract = SuppressionContract(ttl=ttl)

            sup = client.suppressions.create(
                resource_uri=resource_uri,
                recommendation_id=recommendation_id,
                name=suppression_name,
                suppression_contract=suppression_contract
            )

            if rec.suppression_ids:
                rec.suppression_ids.append(sup.suppression_id)
            else:
                rec.suppression_ids = [sup.suppression_id]

    return recs


def enable_recommendations(client, ids=None, resource_group_name=None, recommendation_name=None):
    if recommendation_name is None and ids is None:
        from knack.prompting import prompt_y_n

        if not prompt_y_n("\nAre you sure you want to enable all recommendation?"):
            return None

    recs = _get_recommendations(
        client=client.recommendations,
        ids=ids,
        resource_group_name=resource_group_name,
        recommendation_name=recommendation_name)
    all_sups = list(client.suppressions.list())

    for rec in recs:
        for sup in all_sups:
            try:
                if sup.suppression_id in rec.suppression_ids:
                    result = _parse_recommendation_uri(rec.id)
                    client.suppressions.delete(
                        resource_uri=result['resource_uri'],
                        recommendation_id=result['recommendation_id'],
                        name=sup.name)
            except TypeError:  # when rec.id is already suppressed, rec.suppression_ids is None
                pass
        rec.suppression_ids = None

    return recs


def list_configuration(client):
    return client.list_by_subscription()


def show_configuration(client, resource_group_name=None):
    output = None
    if resource_group_name:
        output = client.list_by_resource_group(resource_group_name)
    else:
        output = client.list_by_subscription()
    # the list is guaranteed to have one element
    return list(output)[0]


def update_configuration(instance, low_cpu_threshold=None,
                         exclude=None, include=None):
    instance.low_cpu_threshold = low_cpu_threshold
    instance.exclude = exclude
    if include:
        instance.exclude = False

    return instance


def _build_filter_string(ids=None, resource_group_name=None, category=None):

    idFilter = None
    if ids:
        idFilter = ' or '.join(["ResourceId eq '{}'".format(id_arg) for id_arg in ids])
    elif resource_group_name:
        idFilter = "ResourceGroup eq '{}'".format(resource_group_name)

    categoryFilter = "Category eq '{}'".format(category) if category else None

    if idFilter:
        if categoryFilter:
            return '({}) and {}'.format(idFilter, categoryFilter)
        return idFilter
    if categoryFilter:
        return categoryFilter

    return None


def _parse_operation_id(location):
    # extract the operation ID from the Location header
    # it is a GUID (i.e. a string of length 36) immediately preceding the api-version query parameter
    end = location.find('?api-version')
    start = end - 36
    operation_id = location[start:end]
    return operation_id


def _parse_recommendation_uri(recommendation_uri):
    resource_uri = recommendation_uri[:recommendation_uri.find("/providers/Microsoft.Advisor/recommendations")]
    rStart = recommendation_uri.find("/recommendations/") + len('/recommendations/')
    # recommendation ID is a GUID (i.e. a string of length 36)
    rEnd = rStart + 36
    recommendation_id = recommendation_uri[rStart:rEnd]
    return {'resource_uri': resource_uri, 'recommendation_id': recommendation_id}


def _generate_recommendations(client):
    from azure.core.exceptions import HttpResponseError

    response = client.generate(cls=_callback)
    location = response.headers['Location']
    operation_id = _parse_operation_id(location)

    try:
        client.get_generate_status(operation_id=operation_id)
    except HttpResponseError as ex:
        # Advisor API returns 204 which is not aligned with ARM guidelines
        # so the SDK will throw an exception that we will have to ignore
        if ex.status_code != 204:
            raise ex


def _set_configuration(client, resource_group_name=None, parameters=None, configuration_name=None):

    if not configuration_name:
        configuration_name = 'default'

    if resource_group_name:
        return client.create_in_resource_group(
            config_contract=parameters,
            resource_group=resource_group_name,
            configuration_name=configuration_name)

    return client.create_in_subscription(config_contract=parameters,
                                         configuration_name=configuration_name)


def _get_recommendations(client, ids=None, resource_group_name=None, recommendation_name=None):
    if ids:
        resource_ids = [_parse_recommendation_uri(id_arg)['resource_uri'] for id_arg in ids]
        recs = list_recommendations(client=client, ids=resource_ids)
        return [r for r in recs if r.id in ids]

    recs = list_recommendations(client=client, resource_group_name=resource_group_name)
    if recommendation_name:
        return [r for r in recs if r.name == recommendation_name]
    return recs


def _callback(*args):
    return args[0].http_response



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=no-self-use, line-too-long, protected-access, too-few-public-methods, unused-argument, unnecessary-pass
from knack.log import get_logger
from knack.util import CLIError

from azure.cli.core.aaz import has_value, register_command, register_command_group, AAZCommandGroup
from ..aaz.latest.disk import Update as _DiskUpdate, GrantAccess as _DiskGrantAccess, Show, UpdatePatch as _UpdatePatch

logger = get_logger(__name__)


class DiskUpdate(_DiskUpdate):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        from azure.cli.core.aaz import AAZStrArg
        args_schema = super()._build_arguments_schema(*args, **kwargs)

        args_schema.disk_access_id._registered = False
        args_schema.disk_encryption_set_id._registered = False

        args_schema.disk_access = AAZStrArg(
            options=["--disk-access"],
            help="Name or ID of the disk access resource for using private endpoints on disks.",
        )
        args_schema.disk_encryption_set = AAZStrArg(
            options=["--disk-encryption-set"],
            help="Name or ID of disk encryption set that is used to encrypt the disk."
        )

        return args_schema

    def pre_instance_update(self, instance):
        from azure.mgmt.core.tools import resource_id, is_valid_resource_id
        from azure.cli.core.commands.client_factory import get_subscription_id

        args = self.ctx.args
        if has_value(args.disk_encryption_set):
            if instance.properties.encryption.type != 'EncryptionAtRestWithCustomerKey' and \
                    has_value(args.encryption_type) and \
                    args.encryption_type != 'EncryptionAtRestWithCustomerKey':
                raise CLIError('usage error: Please set --encryption-type to EncryptionAtRestWithCustomerKey')

            disk_encryption_set = args.disk_encryption_set
            if not is_valid_resource_id(disk_encryption_set.to_serialized_data()):
                disk_encryption_set = resource_id(
                    subscription=get_subscription_id(self.cli_ctx), resource_group=args.resource_group,
                    namespace='Microsoft.Compute', type='diskEncryptionSets', name=disk_encryption_set)

            instance.properties.encryption.disk_encryption_set_id = disk_encryption_set

        if has_value(args.encryption_type):
            if args.encryption_type != 'EncryptionAtRestWithCustomerKey':
                instance.properties.encryption.disk_encryption_set_id = None

        if has_value(args.disk_access):
            disk_access = args.disk_access
            if not is_valid_resource_id(disk_access.to_serialized_data()):
                disk_access = resource_id(
                    subscription=get_subscription_id(self.cli_ctx), resource_group=args.resource_group,
                    namespace='Microsoft.Compute', type='diskAccesses', name=disk_access)
            instance.properties.disk_access_id = disk_access


@register_command_group(
    "disk config",
)
class __CMDGroup(AAZCommandGroup):
    """Manage disk config.
    """
    pass


@register_command(
    "disk config update",
)
class DiskConfigUpdate(_UpdatePatch):
    """Update disk config.

    :example: Update disk size.
        az disk config update --name MyManagedDisk --resource-group MyResourceGroup --size-gb 20
    """
    pass


class DiskGrantAccess(_DiskGrantAccess):
    def pre_operations(self):
        args = self.ctx.args

        disk_info = Show(cli_ctx=self.cli_ctx)(command_args={
            "disk_name": args.disk_name,
            "resource_group": args.resource_group
        })

        if disk_info.get("creation_data", None) and \
                disk_info["creation_data"].get("create_option", None) == "UploadPreparedSecure":
            args.secure_vm_guest_state_sas = True



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=no-self-use, line-too-long, protected-access, too-few-public-methods, unused-argument
from knack.log import get_logger

from azure.cli.core.aaz import AAZStrType
from ..aaz.latest.vm.extension import Show as _VMExtensionShow, Create as _VMExtensionCreate, List as _VMExtensionList

logger = get_logger(__name__)


class VMExtensionShow(_VMExtensionShow):

    class VirtualMachineExtensionsGet(_VMExtensionShow.VirtualMachineExtensionsGet):

        # Override to solve key conflict of _schema_on_200.properties.type when deserializing output
        @classmethod
        def _build_schema_on_200(cls):
            schema = super()._build_schema_on_200()

            del schema.properties._fields['type']
            schema.properties.type = AAZStrType(
                serialized_name="typePropertiesType",
            )
            return schema


class VMExtensionList(_VMExtensionList):

    class VirtualMachineExtensionsList(_VMExtensionList.VirtualMachineExtensionsList):

        # Override to solve key conflict of _schema_on_200.value.Element.properties.type when deserializing output
        @classmethod
        def _build_schema_on_200(cls):
            schema = super()._build_schema_on_200()

            del schema.value.Element.properties._fields['type']
            schema.value.Element.properties.type = AAZStrType(
                serialized_name="typePropertiesType",
            )
            return schema


class VMExtensionCreate(_VMExtensionCreate):

    class VirtualMachineExtensionsCreateOrUpdate(_VMExtensionCreate.VirtualMachineExtensionsCreateOrUpdate):

        # Override to solve key conflict of _schema_on_200.properties.type when deserializing output
        @classmethod
        def _build_schema_on_200_201(cls):
            schema = super()._build_schema_on_200_201()

            del schema.properties._fields['type']
            schema.properties.type = AAZStrType(
                serialized_name="typePropertiesType",
            )
            return schema



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long, too-many-locals
from azure.cli.core.commands import CliCommandType
from azure.cli.command_modules.privatedns._client_factory import cf_privatedns_mgmt_zones
from azure.cli.command_modules.privatedns._format import (transform_privatedns_zone_table_output, transform_privatedns_link_table_output, transform_privatedns_record_set_output, transform_privatedns_record_set_table_output)


def load_command_table(self, _):

    network_privatedns_zone_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.privatedns.operations#PrivateZonesOperations.{}',
        client_factory=cf_privatedns_mgmt_zones
    )

    with self.command_group("network private-dns zone", network_privatedns_zone_sdk) as g:
        from .aaz.latest.network.private_dns.zone import List as PrivateDNSZoneList, Show as PrivateDNSZoneShow
        from .custom import PrivateDNSZoneCreate
        self.command_table["network private-dns zone create"] = PrivateDNSZoneCreate(loader=self)
        self.command_table["network private-dns zone list"] = PrivateDNSZoneList(loader=self, table_transformer=transform_privatedns_zone_table_output)
        self.command_table["network private-dns zone show"] = PrivateDNSZoneShow(loader=self, table_transformer=transform_privatedns_zone_table_output)
        g.custom_command("import", "import_zone")
        g.custom_command("export", "export_zone")

    with self.command_group("network private-dns link vnet"):
        from .aaz.latest.network.private_dns.link.vnet import List as PrivateDNSLinkVNetList, Show as PrivateDNSLinkVNetShow
        from .custom import PrivateDNSLinkVNetCreate
        self.command_table["network private-dns link vnet create"] = PrivateDNSLinkVNetCreate(loader=self)
        self.command_table["network private-dns link vnet list"] = PrivateDNSLinkVNetList(loader=self, table_transformer=transform_privatedns_link_table_output)
        self.command_table["network private-dns link vnet show"] = PrivateDNSLinkVNetShow(loader=self, table_transformer=transform_privatedns_link_table_output)

    from .custom import RecordSetACreate, RecordSetAAAACreate, RecordSetCNAMECreate, RecordSetMXCreate, RecordSetPTRCreate, RecordSetSRVCreate, RecordSetTXTCreate
    self.command_table["network private-dns record-set a create"] = RecordSetACreate(loader=self)
    self.command_table["network private-dns record-set aaaa create"] = RecordSetAAAACreate(loader=self)
    self.command_table["network private-dns record-set cname create"] = RecordSetCNAMECreate(loader=self)
    self.command_table["network private-dns record-set mx create"] = RecordSetMXCreate(loader=self)
    self.command_table["network private-dns record-set ptr create"] = RecordSetPTRCreate(loader=self)
    self.command_table["network private-dns record-set srv create"] = RecordSetSRVCreate(loader=self)
    self.command_table["network private-dns record-set txt create"] = RecordSetTXTCreate(loader=self)

    from .custom import RecordSetADelete, RecordSetAAAADelete, RecordSetCNAMEDelete, RecordSetMXDelete, RecordSetPTRDelete, RecordSetSRVDelete, RecordSetTXTDelete
    self.command_table["network private-dns record-set a delete"] = RecordSetADelete(loader=self)
    self.command_table["network private-dns record-set aaaa delete"] = RecordSetAAAADelete(loader=self)
    self.command_table["network private-dns record-set cname delete"] = RecordSetCNAMEDelete(loader=self)
    self.command_table["network private-dns record-set mx delete"] = RecordSetMXDelete(loader=self)
    self.command_table["network private-dns record-set ptr delete"] = RecordSetPTRDelete(loader=self)
    self.command_table["network private-dns record-set srv delete"] = RecordSetSRVDelete(loader=self)
    self.command_table["network private-dns record-set txt delete"] = RecordSetTXTDelete(loader=self)

    from .aaz.latest.network.private_dns.record_set import List as RecordSetList
    from .custom import RecordSetAList, RecordSetAAAAList, RecordSetCNAMEList, RecordSetMXList, RecordSetPTRList, RecordSetSRVList, RecordSetTXTList
    self.command_table["network private-dns record-set list"] = RecordSetList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set a list"] = RecordSetAList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set aaaa list"] = RecordSetAAAAList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set cname list"] = RecordSetCNAMEList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set mx list"] = RecordSetMXList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set ptr list"] = RecordSetPTRList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set srv list"] = RecordSetSRVList(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set txt list"] = RecordSetTXTList(loader=self, table_transformer=transform_privatedns_record_set_table_output)

    from .custom import RecordSetAShow, RecordSetAAAAShow, RecordSetCNAMEShow, RecordSetMXShow, RecordSetPTRShow, RecordSetSOAShow, RecordSetSRVShow, RecordSetTXTShow
    self.command_table["network private-dns record-set a show"] = RecordSetAShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set aaaa show"] = RecordSetAAAAShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set cname show"] = RecordSetCNAMEShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set mx show"] = RecordSetMXShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set ptr show"] = RecordSetPTRShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set soa show"] = RecordSetSOAShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set srv show"] = RecordSetSRVShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)
    self.command_table["network private-dns record-set txt show"] = RecordSetTXTShow(loader=self, table_transformer=transform_privatedns_record_set_table_output)

    from .custom import RecordSetAUpdate, RecordSetAAAAUpdate, RecordSetCNAMEUpdate, RecordSetMXUpdate, RecordSetPTRUpdate, RecordSetSRVUpdate, RecordSetTXTUpdate
    self.command_table["network private-dns record-set a update"] = RecordSetAUpdate(loader=self)
    self.command_table["network private-dns record-set aaaa update"] = RecordSetAAAAUpdate(loader=self)
    self.command_table["network private-dns record-set cname update"] = RecordSetCNAMEUpdate(loader=self)
    self.command_table["network private-dns record-set mx update"] = RecordSetMXUpdate(loader=self)
    self.command_table["network private-dns record-set ptr update"] = RecordSetPTRUpdate(loader=self)
    self.command_table["network private-dns record-set srv update"] = RecordSetSRVUpdate(loader=self)
    self.command_table["network private-dns record-set txt update"] = RecordSetTXTUpdate(loader=self)

    supported_records = ['a', 'aaaa', 'mx', 'ptr', 'srv', 'txt']
    for record in supported_records:
        with self.command_group('network private-dns record-set {}'.format(record)) as g:
            g.custom_command('add-record', 'add_privatedns_{}_record'.format(record), transform=transform_privatedns_record_set_output)
            g.custom_command('remove-record', 'remove_privatedns_{}_record'.format(record), transform=transform_privatedns_record_set_output)

    with self.command_group('network private-dns record-set soa') as g:
        g.custom_command('update', 'update_privatedns_soa_record', transform=transform_privatedns_record_set_output)

    with self.command_group('network private-dns record-set cname') as g:
        g.custom_command('set-record', 'add_privatedns_cname_record', transform=transform_privatedns_record_set_output)
        g.custom_command('remove-record', 'remove_privatedns_cname_record', transform=transform_privatedns_record_set_output)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from functools import reduce
from azure.cli.core.azclierror import ValidationError

from azure.cli.command_modules.netappfiles._client_factory import (
    volumes_mgmt_client_factory,
    volume_quota_rules_mgmt_client_factory)

from knack.log import get_logger
logger = get_logger(__name__)


def rgetattr(obj, attr, *args):
    """See https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects"""
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return reduce(_getattr, [obj] + attr.split('.'))


# Command validators
def validate_resync_quotarule(cmd, namespace):
    from azure.mgmt.core.tools import parse_resource_id, is_valid_resource_id
    # Get volume params
    netAppVolumeClient = volumes_mgmt_client_factory(cmd.cli_ctx, None)
    netAppVolumeQuotaRuleClient = volume_quota_rules_mgmt_client_factory(cmd.cli_ctx, None)
    logger.debug("ANF validate_resync_quotarule:")
    volume = netAppVolumeClient.get(namespace.resource_group_name, namespace.account_name, namespace.pool_name,
                                    namespace.volume_name)
    logger.debug("ANF validate_resync_quotarule: get volume %s  ", volume.name)
    if volume.data_protection is not None and volume.data_protection.replication is not None:
        endpointType = rgetattr(volume, 'data_protection.replication.endpoint_type')
        remoteVolumeResourceId = rgetattr(volume, 'data_protection.replication.remote_volume_resource_id')
        logger.debug("ANF validate_resync_quotarule: get dataProtection endpointType: %s  ", endpointType)
        if endpointType is not None and endpointType == 'Src':
            logger.debug("ANF validate_resync_quotarule: Get dataprotection volume for volume %s RemoteVolumeResourceId\
                         %s:", namespace.volume_name, remoteVolumeResourceId)
            if is_valid_resource_id(remoteVolumeResourceId):
                resource_parts = parse_resource_id(remoteVolumeResourceId)
                logger.debug("ANF validate_resync_quotarule: Destination volume resourceID parts %s:", resource_parts)
                destGroupName = resource_parts['resource_group']
                destAccountName = resource_parts['name']
                destPoolName = resource_parts['child_name_1']
                destVolumeName = resource_parts['child_name_2']
                logger.debug("ANF validate_resync_quotarule: Get quota rules for volume %s:", destVolumeName)
                try:
                    destVol = netAppVolumeClient.get(destGroupName, destAccountName, destPoolName, destVolumeName)
                    if destVol is not None:
                        destVolumeQuotaRules = netAppVolumeQuotaRuleClient.list_by_volume(destGroupName,
                                                                                          destAccountName, destPoolName,
                                                                                          destVolumeName)
                        rules_list = list(destVolumeQuotaRules)

                except:
                    raise ValidationError("\nDestination volume not found check remote_volume_resource_id: '{}'."
                                          .format(remoteVolumeResourceId))
            else:
                logger.warning("ANF validate_resync_quotarule: The remote_volume_resource_id %s is not a valid\
                               resourceId:", remoteVolumeResourceId)
            if rules_list is not None and not rules_list:
                logger.debug("ANF validate_resync_quotarule: There are no quota rules for volume:")
            else:
                logger.debug("ANF validate_resync_quotarule: There are %s quota rules for this volume, show warning",
                             len(rules_list))
                logger.warning("\nIf any quota rules exists on destination volume they will be overwritten\
with source volume's quota rules.")
    else:
        logger.debug("ANF validate_resync_quotarule: Volume %s is not in a CRR relationship:", namespace.volume_name)
        raise ValueError("\nThis volume is not configured for CRR replication.")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "sig gallery-application version create",
)
class Create(AAZCommand):
    """Create a gallery application version.

    :example: Create a simple gallery Application Version.
        az sig gallery-application version create --gallery-name myGalleryName --application-name myGalleryApplicationName -n 1.0.0 -g myResourceGroup --package-file-link https://{myStorageAccount}.blob.core.windows.net/{myStorageContainer}/{myStorageBlob}  --install-command installCommand  --remove-command removeCommand
    """

    _aaz_info = {
        "version": "2022-01-03",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.compute/galleries/{}/applications/{}/versions/{}", "2022-01-03"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.gallery_application_name = AAZStrArg(
            options=["--application-name", "--gallery-application-name"],
            help="The name of the gallery application.",
            required=True,
        )
        _args_schema.gallery_application_version_name = AAZStrArg(
            options=["-n", "--name", "--version-name", "--gallery-application-version-name"],
            help="The name of the gallery application version.",
            required=True,
        )
        _args_schema.gallery_name = AAZStrArg(
            options=["-r", "--gallery-name"],
            help="Gallery name.",
            required=True,
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.location = AAZResourceLocationArg(
            help="Resource location",
            required=True,
            fmt=AAZResourceLocationArgFormat(
                resource_group_arg="resource_group",
            ),
        )
        _args_schema.install_command = AAZStrArg(
            options=["--install-command"],
            help="Required. The path and arguments to install the gallery application. This is limited to 4096 characters.",
        )
        _args_schema.remove_command = AAZStrArg(
            options=["--remove-command"],
            help="Required. The path and arguments to remove the gallery application. This is limited to 4096 characters.",
        )
        _args_schema.update_command = AAZStrArg(
            options=["--update-command"],
            help="Optional. The path and arguments to update the gallery application. If not present, then update operation will invoke remove command on the previous version and install command on the current version of the gallery application. This is limited to 4096 characters.",
        )
        _args_schema.config_file_name = AAZStrArg(
            options=["--config-file-name"],
            help="Optional. The name to assign the downloaded config file on the VM. This is limited to 4096 characters. If not specified, the config file will be named the Gallery Application name appended with \"_config\".",
        )
        _args_schema.package_file_name = AAZStrArg(
            options=["--package-file-name"],
            help="Optional. The name to assign the downloaded package file on the VM. This is limited to 4096 characters. If not specified, the package file will be named the same as the Gallery Application name.",
        )
        _args_schema.default_file_link = AAZStrArg(
            options=["--default-file-link"],
            help="Optional. The default configuration link of the artifact, must be a readable storage page blob.",
        )
        _args_schema.package_file_link = AAZStrArg(
            options=["--package-file-link"],
            help="Required. The mediaLink of the artifact, must be a readable storage page blob.",
        )
        _args_schema.target_regions = AAZListArg(
            options=["--target-regions"],
            help="The target regions where the Image Version is going to be replicated to. This property is updatable.",
        )
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            help="Space-separated tags: key[=value] [key[=value] ...].",
        )

        target_regions = cls._args_schema.target_regions
        target_regions.Element = AAZObjectArg()

        _element = cls._args_schema.target_regions.Element
        _element.encryption = AAZObjectArg(
            options=["encryption"],
            help="Optional. Allows users to provide customer managed keys for encrypting the OS and data disks in the gallery artifact.",
        )
        cls._build_args_encryption_images_create(_element.encryption)
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the region.",
            required=True,
        )
        _element.regional_replica_count = AAZIntArg(
            options=["regional-replica-count"],
            help="The number of replicas of the Image Version to be created per region. This property is updatable.",
        )
        _element.storage_account_type = AAZStrArg(
            options=["storage-account-type"],
            help="Specifies the storage account type to be used to store the image. This property is not updatable.",
            enum={"Premium_LRS": "Premium_LRS", "Standard_LRS": "Standard_LRS", "Standard_ZRS": "Standard_ZRS"},
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg()

        # define Arg Group "PublishingProfile"

        _args_schema = cls._args_schema
        _args_schema.end_of_life_date = AAZDateTimeArg(
            options=["--end-of-life-date"],
            arg_group="PublishingProfile",
            help="The end of life date of the gallery image version. This property can be used for decommissioning purposes. This property is updatable.",
        )
        _args_schema.exclude_from = AAZBoolArg(
            options=["--exclude-from"],
            arg_group="PublishingProfile",
            help="If set to true, Virtual Machines deployed from the latest version of the Image Definition won't use this Image Version.",
        )
        return cls._args_schema

    _args_encryption_images_create = None

    @classmethod
    def _build_args_encryption_images_create(cls, _schema):
        if cls._args_encryption_images_create is not None:
            _schema.data_disk_images = cls._args_encryption_images_create.data_disk_images
            _schema.os_disk_image = cls._args_encryption_images_create.os_disk_image
            return

        cls._args_encryption_images_create = AAZObjectArg()

        encryption_images_create = cls._args_encryption_images_create
        encryption_images_create.data_disk_images = AAZListArg(
            options=["data-disk-images"],
            help="A list of encryption specifications for data disk images.",
        )
        encryption_images_create.os_disk_image = AAZObjectArg(
            options=["os-disk-image"],
            help="Contains encryption settings for an OS disk image.",
        )

        data_disk_images = cls._args_encryption_images_create.data_disk_images
        data_disk_images.Element = AAZObjectArg()

        _element = cls._args_encryption_images_create.data_disk_images.Element
        _element.disk_encryption_set_id = AAZStrArg(
            options=["disk-encryption-set-id"],
            help="A relative URI containing the resource ID of the disk encryption set.",
        )
        _element.lun = AAZIntArg(
            options=["lun"],
            help="This property specifies the logical unit number of the data disk. This value is used to identify data disks within the Virtual Machine and therefore must be unique for each data disk attached to the Virtual Machine.",
            required=True,
        )

        os_disk_image = cls._args_encryption_images_create.os_disk_image
        os_disk_image.disk_encryption_set_id = AAZStrArg(
            options=["disk-encryption-set-id"],
            help="A relative URI containing the resource ID of the disk encryption set.",
        )
        os_disk_image.security_profile = AAZObjectArg(
            options=["security-profile"],
            help="This property specifies the security profile of an OS disk image.",
        )

        security_profile = cls._args_encryption_images_create.os_disk_image.security_profile
        security_profile.confidential_vm_encryption_type = AAZStrArg(
            options=["confidential-vm-encryption-type"],
            help="confidential VM encryption types",
            enum={"EncryptedVMGuestStateOnlyWithPmk": "EncryptedVMGuestStateOnlyWithPmk", "EncryptedWithCmk": "EncryptedWithCmk", "EncryptedWithPmk": "EncryptedWithPmk"},
        )
        security_profile.secure_vm_disk_encryption_set_id = AAZStrArg(
            options=["secure-vm-disk-encryption-set-id"],
            help="secure VM disk encryption set id",
        )

        _schema.data_disk_images = cls._args_encryption_images_create.data_disk_images
        _schema.os_disk_image = cls._args_encryption_images_create.os_disk_image

    def _execute_operations(self):
        self.pre_operations()
        yield self.GalleryApplicationVersionsCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class GalleryApplicationVersionsCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Compute/galleries/{galleryName}/applications/{galleryApplicationName}/versions/{galleryApplicationVersionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "galleryApplicationName", self.ctx.args.gallery_application_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryApplicationVersionName", self.ctx.args.gallery_application_version_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "galleryName", self.ctx.args.gallery_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2022-01-03",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("location", AAZStrType, ".location", typ_kwargs={"flags": {"required": True}})
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("publishingProfile", AAZObjectType, ".", typ_kwargs={"flags": {"required": True}})

            publishing_profile = _builder.get(".properties.publishingProfile")
            if publishing_profile is not None:
                publishing_profile.set_prop("endOfLifeDate", AAZStrType, ".end_of_life_date")
                publishing_profile.set_prop("excludeFromLatest", AAZBoolType, ".exclude_from")
                publishing_profile.set_prop("manageActions", AAZObjectType)
                publishing_profile.set_prop("settings", AAZObjectType)
                publishing_profile.set_prop("source", AAZObjectType, ".", typ_kwargs={"flags": {"required": True}})
                publishing_profile.set_prop("targetRegions", AAZListType, ".target_regions")

            manage_actions = _builder.get(".properties.publishingProfile.manageActions")
            if manage_actions is not None:
                manage_actions.set_prop("install", AAZStrType, ".install_command", typ_kwargs={"flags": {"required": True}})
                manage_actions.set_prop("remove", AAZStrType, ".remove_command", typ_kwargs={"flags": {"required": True}})
                manage_actions.set_prop("update", AAZStrType, ".update_command")

            settings = _builder.get(".properties.publishingProfile.settings")
            if settings is not None:
                settings.set_prop("configFileName", AAZStrType, ".config_file_name")
                settings.set_prop("packageFileName", AAZStrType, ".package_file_name")

            source = _builder.get(".properties.publishingProfile.source")
            if source is not None:
                source.set_prop("defaultConfigurationLink", AAZStrType, ".default_file_link")
                source.set_prop("mediaLink", AAZStrType, ".package_file_link", typ_kwargs={"flags": {"required": True}})

            target_regions = _builder.get(".properties.publishingProfile.targetRegions")
            if target_regions is not None:
                target_regions.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.publishingProfile.targetRegions[]")
            if _elements is not None:
                _CreateHelper._build_schema_encryption_images_create(_elements.set_prop("encryption", AAZObjectType, ".encryption"))
                _elements.set_prop("name", AAZStrType, ".name", typ_kwargs={"flags": {"required": True}})
                _elements.set_prop("regionalReplicaCount", AAZIntType, ".regional_replica_count")
                _elements.set_prop("storageAccountType", AAZStrType, ".storage_account_type")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _CreateHelper._build_schema_gallery_application_version_read(cls._schema_on_200_201)

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""

    @classmethod
    def _build_schema_encryption_images_create(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("dataDiskImages", AAZListType, ".data_disk_images")
        _builder.set_prop("osDiskImage", AAZObjectType, ".os_disk_image")

        data_disk_images = _builder.get(".dataDiskImages")
        if data_disk_images is not None:
            data_disk_images.set_elements(AAZObjectType, ".")

        _elements = _builder.get(".dataDiskImages[]")
        if _elements is not None:
            _elements.set_prop("diskEncryptionSetId", AAZStrType, ".disk_encryption_set_id")
            _elements.set_prop("lun", AAZIntType, ".lun", typ_kwargs={"flags": {"required": True}})

        os_disk_image = _builder.get(".osDiskImage")
        if os_disk_image is not None:
            os_disk_image.set_prop("diskEncryptionSetId", AAZStrType, ".disk_encryption_set_id")
            os_disk_image.set_prop("securityProfile", AAZObjectType, ".security_profile")

        security_profile = _builder.get(".osDiskImage.securityProfile")
        if security_profile is not None:
            security_profile.set_prop("confidentialVMEncryptionType", AAZStrType, ".confidential_vm_encryption_type")
            security_profile.set_prop("secureVMDiskEncryptionSetId", AAZStrType, ".secure_vm_disk_encryption_set_id")

    _schema_encryption_images_read = None

    @classmethod
    def _build_schema_encryption_images_read(cls, _schema):
        if cls._schema_encryption_images_read is not None:
            _schema.data_disk_images = cls._schema_encryption_images_read.data_disk_images
            _schema.os_disk_image = cls._schema_encryption_images_read.os_disk_image
            return

        cls._schema_encryption_images_read = _schema_encryption_images_read = AAZObjectType()

        encryption_images_read = _schema_encryption_images_read
        encryption_images_read.data_disk_images = AAZListType(
            serialized_name="dataDiskImages",
        )
        encryption_images_read.os_disk_image = AAZObjectType(
            serialized_name="osDiskImage",
        )

        data_disk_images = _schema_encryption_images_read.data_disk_images
        data_disk_images.Element = AAZObjectType()

        _element = _schema_encryption_images_read.data_disk_images.Element
        _element.disk_encryption_set_id = AAZStrType(
            serialized_name="diskEncryptionSetId",
        )
        _element.lun = AAZIntType(
            flags={"required": True},
        )

        os_disk_image = _schema_encryption_images_read.os_disk_image
        os_disk_image.disk_encryption_set_id = AAZStrType(
            serialized_name="diskEncryptionSetId",
        )
        os_disk_image.security_profile = AAZObjectType(
            serialized_name="securityProfile",
        )

        security_profile = _schema_encryption_images_read.os_disk_image.security_profile
        security_profile.confidential_vm_encryption_type = AAZStrType(
            serialized_name="confidentialVMEncryptionType",
        )
        security_profile.secure_vm_disk_encryption_set_id = AAZStrType(
            serialized_name="secureVMDiskEncryptionSetId",
        )

        _schema.data_disk_images = cls._schema_encryption_images_read.data_disk_images
        _schema.os_disk_image = cls._schema_encryption_images_read.os_disk_image

    _schema_gallery_application_version_read = None

    @classmethod
    def _build_schema_gallery_application_version_read(cls, _schema):
        if cls._schema_gallery_application_version_read is not None:
            _schema.id = cls._schema_gallery_application_version_read.id
            _schema.location = cls._schema_gallery_application_version_read.location
            _schema.name = cls._schema_gallery_application_version_read.name
            _schema.properties = cls._schema_gallery_application_version_read.properties
            _schema.tags = cls._schema_gallery_application_version_read.tags
            _schema.type = cls._schema_gallery_application_version_read.type
            return

        cls._schema_gallery_application_version_read = _schema_gallery_application_version_read = AAZObjectType()

        gallery_application_version_read = _schema_gallery_application_version_read
        gallery_application_version_read.id = AAZStrType(
            flags={"read_only": True},
        )
        gallery_application_version_read.location = AAZStrType(
            flags={"required": True},
        )
        gallery_application_version_read.name = AAZStrType(
            flags={"read_only": True},
        )
        gallery_application_version_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        gallery_application_version_read.tags = AAZDictType()
        gallery_application_version_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_gallery_application_version_read.properties
        properties.provisioning_state = AAZStrType(
            serialized_name="provisioningState",
            flags={"read_only": True},
        )
        properties.publishing_profile = AAZObjectType(
            serialized_name="publishingProfile",
            flags={"required": True},
        )
        properties.replication_status = AAZObjectType(
            serialized_name="replicationStatus",
            flags={"read_only": True},
        )

        publishing_profile = _schema_gallery_application_version_read.properties.publishing_profile
        publishing_profile.advanced_settings = AAZDictType(
            serialized_name="advancedSettings",
        )
        publishing_profile.enable_health_check = AAZBoolType(
            serialized_name="enableHealthCheck",
        )
        publishing_profile.end_of_life_date = AAZStrType(
            serialized_name="endOfLifeDate",
        )
        publishing_profile.exclude_from_latest = AAZBoolType(
            serialized_name="excludeFromLatest",
        )
        publishing_profile.manage_actions = AAZObjectType(
            serialized_name="manageActions",
        )
        publishing_profile.published_date = AAZStrType(
            serialized_name="publishedDate",
            flags={"read_only": True},
        )
        publishing_profile.replica_count = AAZIntType(
            serialized_name="replicaCount",
        )
        publishing_profile.replication_mode = AAZStrType(
            serialized_name="replicationMode",
        )
        publishing_profile.settings = AAZObjectType()
        publishing_profile.source = AAZObjectType(
            flags={"required": True},
        )
        publishing_profile.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )
        publishing_profile.target_extended_locations = AAZListType(
            serialized_name="targetExtendedLocations",
        )
        publishing_profile.target_regions = AAZListType(
            serialized_name="targetRegions",
        )

        advanced_settings = _schema_gallery_application_version_read.properties.publishing_profile.advanced_settings
        advanced_settings.Element = AAZStrType()

        manage_actions = _schema_gallery_application_version_read.properties.publishing_profile.manage_actions
        manage_actions.install = AAZStrType(
            flags={"required": True},
        )
        manage_actions.remove = AAZStrType(
            flags={"required": True},
        )
        manage_actions.update = AAZStrType()

        settings = _schema_gallery_application_version_read.properties.publishing_profile.settings
        settings.config_file_name = AAZStrType(
            serialized_name="configFileName",
        )
        settings.package_file_name = AAZStrType(
            serialized_name="packageFileName",
        )

        source = _schema_gallery_application_version_read.properties.publishing_profile.source
        source.default_configuration_link = AAZStrType(
            serialized_name="defaultConfigurationLink",
        )
        source.media_link = AAZStrType(
            serialized_name="mediaLink",
            flags={"required": True},
        )

        target_extended_locations = _schema_gallery_application_version_read.properties.publishing_profile.target_extended_locations
        target_extended_locations.Element = AAZObjectType()

        _element = _schema_gallery_application_version_read.properties.publishing_profile.target_extended_locations.Element
        _element.encryption = AAZObjectType()
        cls._build_schema_encryption_images_read(_element.encryption)
        _element.extended_location = AAZObjectType(
            serialized_name="extendedLocation",
        )
        _element.extended_location_replica_count = AAZIntType(
            serialized_name="extendedLocationReplicaCount",
        )
        _element.name = AAZStrType()
        _element.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )

        extended_location = _schema_gallery_application_version_read.properties.publishing_profile.target_extended_locations.Element.extended_location
        extended_location.name = AAZStrType()
        extended_location.type = AAZStrType()

        target_regions = _schema_gallery_application_version_read.properties.publishing_profile.target_regions
        target_regions.Element = AAZObjectType()

        _element = _schema_gallery_application_version_read.properties.publishing_profile.target_regions.Element
        _element.encryption = AAZObjectType()
        cls._build_schema_encryption_images_read(_element.encryption)
        _element.name = AAZStrType(
            flags={"required": True},
        )
        _element.regional_replica_count = AAZIntType(
            serialized_name="regionalReplicaCount",
        )
        _element.storage_account_type = AAZStrType(
            serialized_name="storageAccountType",
        )

        replication_status = _schema_gallery_application_version_read.properties.replication_status
        replication_status.aggregated_state = AAZStrType(
            serialized_name="aggregatedState",
            flags={"read_only": True},
        )
        replication_status.summary = AAZListType(
            flags={"read_only": True},
        )

        summary = _schema_gallery_application_version_read.properties.replication_status.summary
        summary.Element = AAZObjectType()

        _element = _schema_gallery_application_version_read.properties.replication_status.summary.Element
        _element.details = AAZStrType(
            flags={"read_only": True},
        )
        _element.progress = AAZIntType(
            flags={"read_only": True},
        )
        _element.region = AAZStrType(
            flags={"read_only": True},
        )
        _element.state = AAZStrType(
            flags={"read_only": True},
        )

        tags = _schema_gallery_application_version_read.tags
        tags.Element = AAZStrType()

        _schema.id = cls._schema_gallery_application_version_read.id
        _schema.location = cls._schema_gallery_application_version_read.location
        _schema.name = cls._schema_gallery_application_version_read.name
        _schema.properties = cls._schema_gallery_application_version_read.properties
        _schema.tags = cls._schema_gallery_application_version_read.tags
        _schema.type = cls._schema_gallery_application_version_read.type


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=no-self-use, line-too-long, protected-access, too-few-public-methods, unused-argument
from knack.log import get_logger

from ..aaz.latest.capacity.reservation.group import List as _CapacityReservationGroupList

logger = get_logger(__name__)


class CapacityReservationGroupList(_CapacityReservationGroupList):
    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        from azure.cli.core.aaz import AAZBoolArg
        args_schema = super()._build_arguments_schema(*args, **kwargs)
        args_schema.vm_instance = AAZBoolArg(
            options=['--vm-instance'],
            help="Retrieve the Virtual Machine Instance "
                 "which are associated to capacity reservation group in the response.",
            nullable=True
        )
        args_schema.vmss_instance = AAZBoolArg(
            options=['--vmss-instance'],
            help="Retrieve the ScaleSet VM Instance which are associated to capacity reservation group in the response.",
            nullable=True
        )
        args_schema.expand._registered = False
        return args_schema

    def pre_operations(self):
        from azure.cli.core.aaz import has_value
        args = self.ctx.args
        if args.vm_instance:
            args.expand = "virtualMachines/$ref"
        if args.vmss_instance:
            if has_value(args.expand):
                args.expand = args.expand.to_serialized_data() + ",virtualMachineScaleSetVMs/$ref"
            else:
                args.expand = "virtualMachineScaleSetVMs/$ref"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import sys

from humanfriendly.terminal.spinners import Spinner

BAR_LEN = 70
EMPTY_LINE = ' ' * BAR_LEN


class ProgressViewBase:
    """ a view base for progress reporting """
    def __init__(self, out):
        self.out = out

    def write(self, args):
        """ writes the progress """
        raise NotImplementedError

    def flush(self):
        """ flushes the message out the pipeline"""
        raise NotImplementedError

    def clear(self):
        """ resets the view to neutral """
        pass  # pylint: disable=unnecessary-pass


class ProgressReporter:
    """ generic progress reporter """
    def __init__(self, message='', value=None, total_value=None):
        self.message = message
        self.value = value
        self.total_val = total_value
        self.closed = False

    def add(self, **kwargs):
        """
        adds a progress report
        :param kwargs: dictionary containing 'message', 'total_val', 'value'
        """
        message = kwargs.get('message', self.message)
        total_val = kwargs.get('total_val', self.total_val)
        value = kwargs.get('value', self.value)
        if value and total_val:
            assert 0 <= value <= total_val
            self.closed = value == total_val
        self.total_val = total_val
        self.value = value
        self.message = message

    def report(self):
        """ report the progress """
        percent = self.value / self.total_val if self.value is not None and self.total_val else None
        return {'message': self.message, 'percent': percent}


class ProgressHook:
    """ sends the progress to the view """
    def __init__(self):
        self.reporter = ProgressReporter()
        self.active_progress = None

    def init_progress(self, progress_view):
        """ activate a view """
        self.active_progress = progress_view

    def add(self, **kwargs):
        """ adds a progress report """
        self.reporter.add(**kwargs)
        self.update()

    def update(self):
        """ updates the view with the progress """
        self.active_progress.write(self.reporter.report())
        self.active_progress.flush()

    def stop(self):
        """ if there is an abupt stop before ending """
        self.reporter.closed = True
        self.add(message='Interrupted')
        self.active_progress.clear()

    def begin(self, **kwargs):
        """ start reporting progress """
        kwargs['message'] = kwargs.get('message', 'Starting')
        self.add(**kwargs)
        self.reporter.closed = False

    def end(self, **kwargs):
        """ ending reporting of progress """
        kwargs['message'] = kwargs.get('message', 'Finished')
        self.reporter.closed = True
        self.add(**kwargs)
        self.active_progress.clear()

    def is_running(self):
        """ whether progress is continuing """
        return not self.reporter.closed


class IndeterminateStandardOut(ProgressViewBase):
    """ custom output for progress reporting """
    def __init__(self, out=None, spinner=None):
        super().__init__(
            out if out else sys.stderr)
        self.spinner = spinner

    def write(self, args):
        """
        writes the progress
        :param args: dictionary containing key 'message'
        """
        if self.spinner is None:
            self.spinner = Spinner(  # pylint: disable=no-member
                label='In Progress', stream=self.out, hide_cursor=False)
        msg = args.get('message', 'In Progress')
        try:
            self.spinner.step(label=msg)
        except OSError:
            pass

    def clear(self):
        self.spinner.clear()
        self.out.flush()

    def flush(self):
        self.out.flush()


def _format_value(msg, percent):
    bar_len = BAR_LEN - len(msg) - 1
    completed = int(bar_len * percent)

    message = '\r{}['.format(msg)
    message += ('#' * completed).ljust(bar_len)
    message += ']  {:.4%}'.format(percent)
    return message


class DeterminateStandardOut(ProgressViewBase):
    """ custom output for progress reporting """
    def __init__(self, out=None):
        super().__init__(out if out else sys.stderr)

    def write(self, args):
        """
        writes the progress
        :param args: args is a dictionary containing 'percent', 'message'
        """
        percent = args.get('percent', 0)
        message = args.get('message', '')

        if percent:
            progress = _format_value(message, percent)
            self.out.write(progress)

    def clear(self):
        self.out.write('\n')

    def flush(self):
        self.out.flush()


def get_progress_view(determinant=False, outstream=sys.stderr, spinner=None):
    """ gets your view """
    if determinant:
        return DeterminateStandardOut(out=outstream)
    return IndeterminateStandardOut(out=outstream, spinner=spinner)


class IndeterminateProgressBar:
    """ Define progress bar update view """
    def __init__(self, cli_ctx, message="Running"):
        self.cli_ctx = cli_ctx
        self.message = message
        self.hook = self.cli_ctx.get_progress_controller(
            det=False,
            spinner=Spinner(  # pylint: disable=no-member
                label='Running',
                stream=sys.stderr,
                hide_cursor=False))

    def begin(self):
        self.hook.begin()

    def stop(self):
        self.hook.stop()

    def update_progress(self):
        self.hook.add(message=self.message)

    def update_progress_with_msg(self, message):
        if message != "":
            self.hook.add(message=message)
        else:
            self.hook.add(message=self.message)

    def end(self):
        self.hook.end()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core import AzCommandsLoader
from azure.cli.command_modules.serviceconnector._help import helps  # pylint: disable=unused-import


class MicrosoftServiceConnectorCommandsLoader(AzCommandsLoader):

    def __init__(self, cli_ctx=None):
        from azure.cli.core.commands import CliCommandType
        from azure.cli.command_modules.serviceconnector._client_factory import cf_connection_cl
        connection_custom = CliCommandType(
            operations_tmpl='azure.cli.command_modules.serviceconnector.custom#{}',
            client_factory=cf_connection_cl)
        super().__init__(cli_ctx=cli_ctx, custom_command_type=connection_custom)

    def load_command_table(self, args):
        from azure.cli.command_modules.serviceconnector.commands import load_command_table as load_command_table_manual
        load_command_table_manual(self, args)
        return self.command_table

    def load_arguments(self, command):
        from azure.cli.command_modules.serviceconnector._params import load_arguments as load_arguments_manual
        load_arguments_manual(self, command)


COMMAND_LOADER_CLS = MicrosoftServiceConnectorCommandsLoader



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "cdn custom-domain create",
)
class Create(AAZCommand):
    """Create a new custom domain within an endpoint.

    :example: Create a custom domain with resource name customdomain1 within an endpoint and profile.
        az cdn custom-domain create -g group --endpoint-name endpoint --profile-name profile -n customdomain1 --hostname www.example.com
    """

    _aaz_info = {
        "version": "2025-06-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.cdn/profiles/{}/endpoints/{}/customdomains/{}", "2025-06-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.custom_domain_name = AAZStrArg(
            options=["-n", "--name", "--custom-domain-name"],
            help="Name of the custom domain within an endpoint.",
            required=True,
        )
        _args_schema.endpoint_name = AAZStrArg(
            options=["--endpoint-name"],
            help="Name of the endpoint under the profile which is unique globally.",
            required=True,
        )
        _args_schema.profile_name = AAZStrArg(
            options=["--profile-name"],
            help="Name of the CDN profile which is unique within the resource group.",
            required=True,
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.hostname = AAZStrArg(
            options=["--hostname"],
            arg_group="Properties",
            help="The host name of the custom domain. Must be a domain name.",
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.CustomDomainsCreate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class CustomDomainsCreate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Cdn/profiles/{profileName}/endpoints/{endpointName}/customDomains/{customDomainName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "customDomainName", self.ctx.args.custom_domain_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "endpointName", self.ctx.args.endpoint_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "profileName", self.ctx.args.profile_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-06-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("hostName", AAZStrType, ".hostname", typ_kwargs={"flags": {"required": True}})

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _CreateHelper._build_schema_custom_domain_read(cls._schema_on_200_201)

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""

    _schema_custom_domain_read = None

    @classmethod
    def _build_schema_custom_domain_read(cls, _schema):
        if cls._schema_custom_domain_read is not None:
            _schema.id = cls._schema_custom_domain_read.id
            _schema.name = cls._schema_custom_domain_read.name
            _schema.properties = cls._schema_custom_domain_read.properties
            _schema.system_data = cls._schema_custom_domain_read.system_data
            _schema.type = cls._schema_custom_domain_read.type
            return

        cls._schema_custom_domain_read = _schema_custom_domain_read = AAZObjectType()

        custom_domain_read = _schema_custom_domain_read
        custom_domain_read.id = AAZStrType(
            flags={"read_only": True},
        )
        custom_domain_read.name = AAZStrType(
            flags={"read_only": True},
        )
        custom_domain_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        custom_domain_read.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        custom_domain_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_custom_domain_read.properties
        properties.custom_https_parameters = AAZObjectType(
            serialized_name="customHttpsParameters",
        )
        properties.custom_https_provisioning_state = AAZStrType(
            serialized_name="customHttpsProvisioningState",
            flags={"read_only": True},
        )
        properties.custom_https_provisioning_substate = AAZStrType(
            serialized_name="customHttpsProvisioningSubstate",
            flags={"read_only": True},
        )
        properties.host_name = AAZStrType(
            serialized_name="hostName",
            flags={"required": True},
        )
        properties.provisioning_state = AAZStrType(
            serialized_name="provisioningState",
            flags={"read_only": True},
        )
        properties.resource_state = AAZStrType(
            serialized_name="resourceState",
            flags={"read_only": True},
        )
        properties.validation_data = AAZStrType(
            serialized_name="validationData",
        )

        custom_https_parameters = _schema_custom_domain_read.properties.custom_https_parameters
        custom_https_parameters.certificate_source = AAZStrType(
            serialized_name="certificateSource",
            flags={"required": True},
        )
        custom_https_parameters.minimum_tls_version = AAZStrType(
            serialized_name="minimumTlsVersion",
        )
        custom_https_parameters.protocol_type = AAZStrType(
            serialized_name="protocolType",
            flags={"required": True},
        )

        disc_azure_key_vault = _schema_custom_domain_read.properties.custom_https_parameters.discriminate_by("certificate_source", "AzureKeyVault")
        disc_azure_key_vault.certificate_source_parameters = AAZObjectType(
            serialized_name="certificateSourceParameters",
            flags={"required": True},
        )

        certificate_source_parameters = _schema_custom_domain_read.properties.custom_https_parameters.discriminate_by("certificate_source", "AzureKeyVault").certificate_source_parameters
        certificate_source_parameters.delete_rule = AAZStrType(
            serialized_name="deleteRule",
            flags={"required": True},
        )
        certificate_source_parameters.resource_group_name = AAZStrType(
            serialized_name="resourceGroupName",
            flags={"required": True},
        )
        certificate_source_parameters.secret_name = AAZStrType(
            serialized_name="secretName",
            flags={"required": True},
        )
        certificate_source_parameters.secret_version = AAZStrType(
            serialized_name="secretVersion",
        )
        certificate_source_parameters.subscription_id = AAZStrType(
            serialized_name="subscriptionId",
            flags={"required": True},
        )
        certificate_source_parameters.type_name = AAZStrType(
            serialized_name="typeName",
            flags={"required": True},
        )
        certificate_source_parameters.update_rule = AAZStrType(
            serialized_name="updateRule",
            flags={"required": True},
        )
        certificate_source_parameters.vault_name = AAZStrType(
            serialized_name="vaultName",
            flags={"required": True},
        )

        disc_cdn = _schema_custom_domain_read.properties.custom_https_parameters.discriminate_by("certificate_source", "Cdn")
        disc_cdn.certificate_source_parameters = AAZObjectType(
            serialized_name="certificateSourceParameters",
            flags={"required": True},
        )

        certificate_source_parameters = _schema_custom_domain_read.properties.custom_https_parameters.discriminate_by("certificate_source", "Cdn").certificate_source_parameters
        certificate_source_parameters.certificate_type = AAZStrType(
            serialized_name="certificateType",
            flags={"required": True},
        )
        certificate_source_parameters.type_name = AAZStrType(
            serialized_name="typeName",
            flags={"required": True},
        )

        system_data = _schema_custom_domain_read.system_data
        system_data.created_at = AAZStrType(
            serialized_name="createdAt",
        )
        system_data.created_by = AAZStrType(
            serialized_name="createdBy",
        )
        system_data.created_by_type = AAZStrType(
            serialized_name="createdByType",
        )
        system_data.last_modified_at = AAZStrType(
            serialized_name="lastModifiedAt",
        )
        system_data.last_modified_by = AAZStrType(
            serialized_name="lastModifiedBy",
        )
        system_data.last_modified_by_type = AAZStrType(
            serialized_name="lastModifiedByType",
        )

        _schema.id = cls._schema_custom_domain_read.id
        _schema.name = cls._schema_custom_domain_read.name
        _schema.properties = cls._schema_custom_domain_read.properties
        _schema.system_data = cls._schema_custom_domain_read.system_data
        _schema.type = cls._schema_custom_domain_read.type


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from enum import Enum
import json
from knack.log import get_logger
from azure.cli.core.util import shell_safe_json_parse
from azure.cli.core.azclierror import InvalidArgumentValueError, ValidationError
from collections import namedtuple
from ._models import KeyValue
from ._constants import FeatureFlagConstants

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=line-too-long
# pylint: disable=too-many-branches

logger = get_logger(__name__)

# Feature Flag Models #


class FeatureState(Enum):
    OFF = 1
    ON = 2
    CONDITIONAL = 3


class FeatureQueryFields(Enum):
    KEY = 0x001
    LABEL = 0x002
    NAME = 0x004
    LAST_MODIFIED = 0x020
    LOCKED = 0x040
    STATE = 0x100
    DESCRIPTION = 0x200
    CONDITIONS = 0x400
    ALL = KEY | LABEL | NAME | LAST_MODIFIED | LOCKED | STATE | DESCRIPTION | CONDITIONS


class FeatureFlagVersion(Enum):
    V1 = 1
    V2 = 2


class FeatureFlagValue:
    """
    Schema of Value inside KeyValue when key is a Feature Flag.

    :ivar str id:
        ID (name) of the feature.
    :ivar str description:
        Description of Feature Flag
    :ivar bool enabled:
        Represents if the Feature flag is On/Off/Conditionally On
    :ivar str display_name:
        A display name for the feature to use for display rather than the ID.
    :ivar dict {string, FeatureFilter[]} conditions:
        Dictionary that contains client_filters List
    :ivar FeatureAllocation allocation:
        Determines how variants should be allocated for the feature to various users.
    :ivar list FeatureVariant variants:
        The list of variants defined for this feature.
    :ivar dict telemetry:
        The declaration of options used to configure telemetry for this feature.
    """

    def __init__(
        self,
        id_,
        description=None,
        enabled=None,
        conditions=None,
        display_name=None,
        variants=None,
        allocation=None,
        telemetry=None,
    ):

        self.id = id_
        self.description = description
        self.enabled = enabled
        self.conditions = conditions
        self.allocation = allocation
        self.variants = variants
        self.telemetry = telemetry
        self.display_name = display_name

    def __repr__(self):
        featureflagvalue = {
            FeatureFlagConstants.ID: self.id,
            FeatureFlagConstants.DESCRIPTION: self.description,
            FeatureFlagConstants.ENABLED: self.enabled,
            FeatureFlagConstants.CONDITIONS: custom_serialize_conditions(self.conditions),
            FeatureFlagConstants.DISPLAY_NAME: self.display_name,
            FeatureFlagConstants.ALLOCATION: custom_serialize_allocation(self.allocation),
            FeatureFlagConstants.VARIANTS: custom_serialize_variants(self.variants),
            FeatureFlagConstants.TELEMETRY: custom_serialize_telemetry(self.telemetry)
        }
        return json.dumps(featureflagvalue, indent=2, ensure_ascii=False)


class FeatureFlag:
    """
    Feature Flag schema as displayed to the user.

    :ivar str name:
        Name (ID) of the feature flag.
    :ivar str key:
        Key of the feature flag.
    :ivar str label:
        Label of the feature flag.
    :ivar str state:
        Represents if the Feature flag is On/Off/Conditionally On
    :ivar bool enabled:
        Represents if the Feature flag is On/Off
    :ivar str description:
        Description of Feature Flag
    :ivar bool locked:
        Represents whether the feature flag is locked.
    :ivar str display_name:
        Display name of the feature flag.
    :ivar str last_modified:
        A str representation of the datetime object representing the last time the feature flag was modified.
    :ivar str etag:
        The ETag contains a value that you can use to perform operations.
    :ivar dict {string, FeatureFilter[]} conditions:
        Dictionary that contains client_filters List
    :ivar dict allocation:
        Determines how variants should be allocated for the feature to various users.
    :ivar list FeatureVariant variants:
        The list of variants defined for this feature.
    :ivar dict {string, object} telemetry:
        The declaration of options used to configure telemetry for this feature.
    :ivar dict tags:
        Tags associated with the feature flag.
    """

    def __init__(
        self,
        name,
        key=None,
        label=None,
        state=None,
        enabled=None,
        description=None,
        conditions=None,
        locked=None,
        display_name=None,
        last_modified=None,
        allocation=None,
        variants=None,
        telemetry=None,
        tags=None
    ):
        self.name = name
        self.key = key
        self.label = label
        self.enabled = enabled
        self.state = state.name.lower()
        self.description = description
        self.conditions = conditions
        self.last_modified = last_modified
        self.locked = locked
        self.allocation = allocation
        self.variants = variants
        self.telemetry = telemetry
        self.display_name = display_name
        self.tags = tags

    def __repr__(self):
        featureflag = {
            "Feature Name": self.name,
            "Key": self.key,
            "Label": self.label,
            "State": self.state,
            "Enabled": self.enabled,
            "Locked": self.locked,
            "Description": self.description,
            "Last Modified": self.last_modified,
            "Conditions": custom_serialize_conditions(self.conditions),
            "Allocation": custom_serialize_allocation(self.allocation),
            "Variants": custom_serialize_variants(self.variants),
            "Telemetry": custom_serialize_telemetry(self.telemetry),
            "Display Name": self.display_name,
            "Tags": self.tags
        }

        return json.dumps(featureflag, indent=2, ensure_ascii=False)


class FeatureFilter:
    """
    Feature filters class.

    :ivar str name:
        Name of the filter
    :ivar dict parameters:
        Name-Value pairs of parameters
    """

    def __init__(self, name, parameters=None):
        self.name = name
        self.parameters = parameters

    def __repr__(self):
        featurefilter = {
            FeatureFlagConstants.NAME: self.name,
            FeatureFlagConstants.FILTER_PARAMETERS: self.parameters,
        }
        return json.dumps(featurefilter, indent=2, ensure_ascii=False)


class FeatureVariant:
    """
    Feature variants class.

    :ivar str name:
        Name of the variant
    :ivar str configuration_Value:
        The configuration value of the variant
    :ivar str Status_Overide:
        Overrides the enabled state of the feature if the given variant is assigned.
        Does not override the state if value is None
    """

    def __init__(self, name, configuration_value=None, status_override=None):
        self.name = name
        self.configuration_value = configuration_value
        self.status_override = status_override

    @classmethod
    def convert_from_dict(cls, variant_dict):
        """
        Convert dict to FeatureVariant object

        Args:
            dictionary - {string, Any}

        Return:
            FeatureVariant object
        """
        name = variant_dict.get(FeatureFlagConstants.NAME, None)
        if not name:
            raise ValidationError(
                "Feature variant must contain required '%s' attribute: \n%s"
                % (FeatureFlagConstants.NAME, json.dumps(variant_dict, indent=2, ensure_ascii=False))
            )

        configuration_value = variant_dict.get(
            FeatureFlagConstants.VARIANT_CONFIGURATION_VALUE, None
        )
        status_override = variant_dict.get(
            FeatureFlagConstants.VARIANT_STATUS_OVERRIDE, None
        )
        return cls(
            name=name,
            configuration_value=configuration_value,
            status_override=status_override,
        )

    def __repr__(self):
        featurevariant = {
            FeatureFlagConstants.NAME: self.name,
            FeatureFlagConstants.VARIANT_CONFIGURATION_VALUE: self.configuration_value,
            FeatureFlagConstants.VARIANT_STATUS_OVERRIDE: self.status_override
        }

        return json.dumps(featurevariant, indent=2, ensure_ascii=False)


class FeatureUserAllocation:
    """
    Feature user allocation class.

    :ivar str variant:
        The name of the variant to use if the user allocation matches the current user.
    :ivar list users:
        Collection of users where if any match the current user, the variant specified in the user allocation is used.
    """

    def __init__(self, variant, users):
        self.variant = variant
        self.users = users

    @classmethod
    def convert_from_dict(cls, user_allocation_dict):
        """
        Convert dict to FeatureUserAllocation object

        Args:
            dictionary - {string, Any}

        Return:
            FeatureUserAllocation object
        """
        variant = user_allocation_dict.get(FeatureFlagConstants.VARIANT, None)
        users = user_allocation_dict.get(FeatureFlagConstants.USERS, None)
        if not variant or not users:
            raise ValidationError(
                "User variant allocation must contain required '%s' and '%s' attributes: \n%s"
                % (FeatureFlagConstants.VARIANT,
                   FeatureFlagConstants.USERS,
                   json.dumps(user_allocation_dict, indent=2, ensure_ascii=False))
            )

        return cls(variant=variant, users=users)

    def __repr__(self):
        feature_user_allocation = {
            FeatureFlagConstants.VARIANT: self.variant,
            FeatureFlagConstants.USERS: self.users,
        }

        return json.dumps(feature_user_allocation, indent=2, ensure_ascii=False)


class FeatureGroupAllocation:
    """
    Feature group allocation class.

    :ivar str variant:
        The name of the variant to use
        if the group allocation matches a group the current user is in.
    :ivar list groups:
        Collection of groups where if the current user is in any of these groups,
        the variant specified in the group allocation is used.
    """

    def __init__(self, variant, groups):
        self.variant = variant
        self.groups = groups

    @classmethod
    def convert_from_dict(cls, group_allocation_dict):
        """
        Convert dict to FeatureGroupAllocation object

        Args:
            dictionary - {string, Any}

        Return:
            FeatureGroupAllocation object
        """
        variant = group_allocation_dict.get(FeatureFlagConstants.VARIANT, None)
        groups = group_allocation_dict.get(FeatureFlagConstants.GROUPS, None)
        if not variant or not groups:
            raise ValidationError(
                "Group variant allocation must contain required '%s' and '%s' attributes: \n%s"
                % (FeatureFlagConstants.VARIANT,
                   FeatureFlagConstants.GROUPS,
                   json.dumps(group_allocation_dict, indent=2, ensure_ascii=False))
            )

        return cls(variant=variant, groups=groups)

    def __repr__(self):
        feature_group_allocation = {
            FeatureFlagConstants.VARIANT: self.variant,
            FeatureFlagConstants.GROUPS: self.groups,
        }

        return json.dumps(feature_group_allocation, indent=2, ensure_ascii=False)


class FeaturePercentileAllocation:
    """
    Feature percentile allocation class.

    :ivar str variant:
        The name of the variant to use if the calculated percentile for the current user falls in the provided range.
    :ivar number from:
        The lower end of the percentage range for which this variant will be used.
    :ivar number to:
        The upper end of the percentage range for which this variant will be used.
    """

    def __init__(self, variant, from_, to):
        self.variant = variant
        setattr(self, "from", from_)
        self.to = to

    @classmethod
    def convert_from_dict(cls, percentile_allocation_dict):
        """
        Convert dict to FeaturePercentileAllocation object

        Args:
            dictionary - {string, Any}

        Return:
            FeaturePercentileAllocation object
        """

        variant = percentile_allocation_dict.get(FeatureFlagConstants.VARIANT, None)
        percentile_from = percentile_allocation_dict.get(FeatureFlagConstants.FROM, None)
        percentile_to = percentile_allocation_dict.get(FeatureFlagConstants.TO, None)
        if not variant or percentile_from is None or percentile_to is None:
            raise ValidationError(
                "Percentile allocation must contain required '%s', '%s' and '%s' attributes: \n%s"
                % (FeatureFlagConstants.VARIANT,
                   FeatureFlagConstants.TO,
                   FeatureFlagConstants.FROM,
                   json.dumps(percentile_allocation_dict, indent=2, ensure_ascii=False))
            )

        if not isinstance(percentile_from, int) or not isinstance(percentile_to, int):
            raise ValidationError(
                "Percentile allocation '%s' and '%s' must be integers: \n%s"
                % (FeatureFlagConstants.FROM,
                   FeatureFlagConstants.TO,
                   json.dumps(percentile_allocation_dict, indent=2, ensure_ascii=False))
            )

        if (
            percentile_from < 0 or
            percentile_from > 100 or
            percentile_to < 0 or
            percentile_to > 100
        ):
            raise ValidationError(
                "Percentile allocation '%s' and '%s' must be between 0 and 100: \n%s"
                % (FeatureFlagConstants.TO,
                   FeatureFlagConstants.FROM,
                   json.dumps(percentile_allocation_dict, indent=2, ensure_ascii=False))
            )

        if percentile_from >= percentile_to:
            raise ValidationError(
                "Percentile allocation '%s' must be less than '%s': \n%s"
                % (FeatureFlagConstants.FROM,
                   FeatureFlagConstants.TO,
                   json.dumps(percentile_allocation_dict, indent=2, ensure_ascii=False))
            )

        return cls(variant=variant, from_=percentile_from, to=percentile_to)

    def __repr__(self):
        feature_percentile_allocation = {
            FeatureFlagConstants.VARIANT: self.variant,
            FeatureFlagConstants.FROM: getattr(self, "from"),
            FeatureFlagConstants.TO: self.to,
        }

        return json.dumps(feature_percentile_allocation, indent=2, ensure_ascii=False)


class FeatureAllocation:
    """
    Feature allocation class.

    :ivar list FeatureUserAllocation user:
        Determines how variants should be allocated for the feature to various users.
    :ivar list FeatureGroupAllocation group:
        Determines how variants should be allocated for the feature to various groups.
    :ivar list FeaturePercentileAllocation percentile:
        Determines how variants should be allocated for the feature to various users based on percentile.
    :ivar str default_when_enabled:
        Specifies which variant should be used
        when the feature is considered enabled and no other allocation rules are applicable.
    :ivar str default_when_disabled:
        Specifies which variant should be used when the feature is considered disabled.
    :ivar str seed:
        The value percentile calculations are based on.
        The calculated percentile is consistent across features for a given user if the same nonempty seed is used.
    """

    def __init__(
        self,
        user=None,
        group=None,
        percentile=None,
        default_when_enabled=None,
        default_when_disabled=None,
        seed=None
    ):
        self.user = user
        self.group = group
        self.percentile = percentile
        self.default_when_enabled = default_when_enabled
        self.default_when_disabled = default_when_disabled
        self.seed = seed

    @classmethod
    def convert_from_dict(cls, allocation_dict):
        """
        Convert dict to FeatureAllocation object

        Args:
            dict {string, Any}

        Return:
            FeatureAllocation object
        """
        default_when_disabled = allocation_dict.get(
            FeatureFlagConstants.DEFAULT_WHEN_DISABLED, None
        )
        default_when_enabled = allocation_dict.get(
            FeatureFlagConstants.DEFAULT_WHEN_ENABLED, None
        )
        seed = allocation_dict.get(
            FeatureFlagConstants.SEED, None
        )

        allocation = cls(
            default_when_enabled=default_when_enabled,
            default_when_disabled=default_when_disabled,
            seed=seed
        )

        allocation_user = allocation_dict.get(FeatureFlagConstants.USER, None)
        # Convert all users to FeatureUserAllocation object
        if allocation_user:
            allocation_user_list = []
            for user in allocation_user:
                if user:
                    feature_user_allocation = FeatureUserAllocation.convert_from_dict(user)
                    allocation_user_list.append(feature_user_allocation)

            allocation.user = allocation_user_list

        # Convert all groups to FeatureGroupAllocation object
        allocation_group = allocation_dict.get(FeatureFlagConstants.GROUP, None)
        if allocation_group:
            allocation_group_list = []
            for group in allocation_group:
                if group:
                    feature_group_allocation = FeatureGroupAllocation.convert_from_dict(group)
                    allocation_group_list.append(feature_group_allocation)

            allocation.group = allocation_group_list

        # Convert all percentile to FeatureAllocationPercentile object
        allocation_percentile = allocation_dict.get(
            FeatureFlagConstants.PERCENTILE, None
        )
        if allocation_percentile:
            allocation_percentile_list = []
            for percentile in allocation_percentile:
                if percentile:
                    feature_percentile_allocation = FeaturePercentileAllocation.convert_from_dict(percentile)
                    allocation_percentile_list.append(feature_percentile_allocation)

            allocation.percentile = allocation_percentile_list

        return allocation

    def __repr__(self):
        featureallocation = {
            FeatureFlagConstants.USER: list(self.user),
            FeatureFlagConstants.GROUP: list(self.group),
            FeatureFlagConstants.PERCENTILE: list(self.percentile),
            FeatureFlagConstants.DEFAULT_WHEN_ENABLED: self.default_when_enabled,
            FeatureFlagConstants.DEFAULT_WHEN_DISABLED: self.default_when_disabled,
            FeatureFlagConstants.SEED: self.seed
        }

        return json.dumps(featureallocation, indent=2, ensure_ascii=False)


class FeatureTelemetry:
    """
    Feature telemtry class.

    :ivar bool enabled:
        Indicates if telemetry is enabled.
    :ivar dict metadata:
        A container for metadata that should be bundled with flag telemetry.
    """
    def __init__(self, enabled=False, metadata=None):
        self.enabled = enabled
        self.metadata = metadata

    @classmethod
    def convert_from_dict(cls, feature_telemetry_dict, feature_name):
        """
        Convert dict to FeatureTelemetry object

        Args:
            dictionary - {string, Any}

        Return:
            FeatureTelemetry object
        """
        enabled = None
        telemetry_enabled = feature_telemetry_dict.get(FeatureFlagConstants.ENABLED, None)
        if telemetry_enabled is not None:
            enabled = convert_string_to_bool(telemetry_enabled, feature_name)

        metadata = feature_telemetry_dict.get(FeatureFlagConstants.METADATA, None)

        return cls(enabled=enabled, metadata=metadata)

    def __repr__(self):
        feature_telemetry = {
            FeatureFlagConstants.ENABLED: self.enabled,
            FeatureFlagConstants.METADATA: self.metadata,
        }

        return json.dumps(feature_telemetry, indent=2, ensure_ascii=False)


# Feature Flag Helper Functions #
def custom_serialize_conditions(conditions_dict):
    """
    Helper Function to serialize Conditions

    Args:
        conditions_dict - Dictionary of {str, Any}

    Return:
        JSON serializable Dictionary
    """
    featurefilterdict = {}

    for key, value in conditions_dict.items():
        if key == FeatureFlagConstants.CLIENT_FILTERS:
            if value is not None:
                featurefilterdict[key] = [
                    feature_filter.__dict__ for feature_filter in value
                ]
        else:
            featurefilterdict[key] = value

    return featurefilterdict


def custom_serialize_allocation(allocation):
    """
    Helper Function to serialize Allocation

    Args:
        allocation_dict - FeatureAllocation object

    Return:
        JSON serializable Dictionary
    """
    featureallocationdict = {}

    if allocation.user:
        featureallocationdict[FeatureFlagConstants.USER] = [
            user.__dict__ for user in allocation.user
        ]

    if allocation.group:
        featureallocationdict[FeatureFlagConstants.GROUP] = [
            group.__dict__ for group in allocation.group
        ]

    if allocation.percentile:
        featureallocationdict[FeatureFlagConstants.PERCENTILE] = [
            percentile.__dict__ for percentile in allocation.percentile
        ]

    if allocation.default_when_enabled:
        featureallocationdict[FeatureFlagConstants.DEFAULT_WHEN_ENABLED] = allocation.default_when_enabled

    if allocation.default_when_disabled:
        featureallocationdict[FeatureFlagConstants.DEFAULT_WHEN_DISABLED] = allocation.default_when_disabled

    if allocation.seed:
        featureallocationdict[FeatureFlagConstants.SEED] = allocation.seed

    return featureallocationdict


def custom_serialize_variants(variants_list):
    """
    Helper Function to serialize Variants

    Args:
        variants_list - List of FeatureVariant objects

    Return:
        JSON serializable List
    """
    featurevariants = []
    for variant in variants_list:
        variant_dict = {}
        variant_dict[FeatureFlagConstants.NAME] = variant.name
        variant_dict[FeatureFlagConstants.VARIANT_CONFIGURATION_VALUE] = variant.configuration_value

        if variant.status_override:
            variant_dict[FeatureFlagConstants.VARIANT_STATUS_OVERRIDE] = variant.status_override

        featurevariants.append(variant_dict)
    return featurevariants


def custom_serialize_telemetry(telemetry):
    """
    Helper Function to serialize Telemetry

    Args:
        telemetry - FeatureTelemetry object

    Return:
        JSON serializable dictionary

    """

    feature_telemetry_dict = {}

    if telemetry.enabled is not None:
        feature_telemetry_dict[FeatureFlagConstants.ENABLED] = telemetry.enabled

    if telemetry.metadata:
        feature_telemetry_dict[FeatureFlagConstants.METADATA] = telemetry.metadata

    return feature_telemetry_dict


def map_featureflag_to_keyvalue(featureflag):
    """
    Helper Function to convert FeatureFlag object to KeyValue object

    Args:
        featureflag - FeatureFlag object to be converted

    Return:
        KeyValue object
    """
    try:
        enabled = False
        if featureflag.state in ("on", "conditional"):
            enabled = True

        feature_flag_value = FeatureFlagValue(
            id_=featureflag.name,
            description=featureflag.description,
            enabled=enabled,
            conditions=featureflag.conditions,
            display_name=featureflag.display_name,
            allocation=featureflag.allocation,
            variants=featureflag.variants,
            telemetry=featureflag.telemetry,
        )

        set_kv = KeyValue(
            key=featureflag.key,
            label=featureflag.label,
            value=json.dumps(
                feature_flag_value, default=lambda o: {k: v for k, v in o.__dict__.items() if v is not None}, ensure_ascii=False
            ),
            content_type=FeatureFlagConstants.FEATURE_FLAG_CONTENT_TYPE,
            tags={},
        )

        set_kv.locked = featureflag.locked
        set_kv.last_modified = featureflag.last_modified

    except ValueError as exception:
        error_msg = "Exception while converting feature flag to key value: {0}\n{1}".format(featureflag.key, exception)
        raise ValueError(error_msg)

    except Exception as exception:
        error_msg = "Exception while converting feature flag to key value: {0}\n{1}".format(featureflag.key, exception)
        raise Exception(error_msg)  # pylint: disable=broad-exception-raised

    return set_kv


def map_keyvalue_to_featureflag(keyvalue, show_all_details=True, hide_enabled=True):
    """
    Helper Function to convert KeyValue object to FeatureFlag object for display

    Args:
        keyvalue - KeyValue object to be converted
        show_all_details - Boolean for controlling whether we want to display "Conditions", "Allocation", "Variants", "Telemetry" or not
        hide_enabled - Boolean for controlling whether we want to hide the "Enabled" field or not
    Return:
        FeatureFlag object
    """

    feature_flag_value = map_keyvalue_to_featureflagvalue(keyvalue)
    feature_name = feature_flag_value.id
    state = FeatureState.OFF
    if feature_flag_value.enabled:
        state = FeatureState.ON

    conditions = feature_flag_value.conditions
    filters = None

    # if conditions["client_filters"] list is not empty, make state conditional
    if conditions and FeatureFlagConstants.CLIENT_FILTERS in conditions:
        filters = conditions[FeatureFlagConstants.CLIENT_FILTERS]

    if (
        filters or
        feature_flag_value.allocation or
        feature_flag_value.variants or
        feature_flag_value.telemetry
    ) and state == FeatureState.ON:
        state = FeatureState.CONDITIONAL

    feature_flag = FeatureFlag(
        name=feature_name,
        key=keyvalue.key,
        label=keyvalue.label,
        state=state,
        enabled=feature_flag_value.enabled,
        description=feature_flag_value.description,
        conditions=conditions,
        locked=keyvalue.locked,
        display_name=feature_flag_value.display_name,
        last_modified=keyvalue.last_modified,
        allocation=feature_flag_value.allocation,
        variants=feature_flag_value.variants,
        telemetry=feature_flag_value.telemetry,
        tags=keyvalue.tags
    )

    # By Default, we will try to show conditions unless the user has
    # specifically filtered them using --fields arg.
    # But in some operations like 'Delete feature', we don't want
    # to display all the conditions as a result of delete operation
    if not show_all_details:
        del feature_flag.conditions
        del feature_flag.allocation
        del feature_flag.variants
        del feature_flag.telemetry

    if hide_enabled:
        del feature_flag.enabled
    else:
        del feature_flag.state
    return feature_flag


def map_keyvalue_to_featureflagvalue(keyvalue):
    """
    Helper Function to convert value string to a valid FeatureFlagValue.
    Throws Exception if value is an invalid JSON.

    Args:
        keyvalue - KeyValue object

    Return:
        Valid FeatureFlagValue object
    """

    try:
        # Make sure value string is a valid json
        feature_flag_dict = shell_safe_json_parse(keyvalue.value)
        feature_flag_version = FeatureFlagVersion.V1  # Default to v1 feature flags

        # Make sure value json has all the fields we support in the backend
        valid_fields = {
            FeatureFlagConstants.ID,
            FeatureFlagConstants.DESCRIPTION,
            FeatureFlagConstants.ENABLED,
            FeatureFlagConstants.CONDITIONS,
            FeatureFlagConstants.ALLOCATION,
            FeatureFlagConstants.VARIANTS,
            FeatureFlagConstants.TELEMETRY,
            FeatureFlagConstants.DISPLAY_NAME
        }
        if valid_fields.union(feature_flag_dict.keys()) != valid_fields:
            logger.debug(
                "'%s' feature flag is missing required values or it contains ",
                keyvalue.key +
                "unsupported values. Unsupported values will be ignored.\n",
            )

        # check if feature flag is version v1 or v2.
        # We always default values for v1 feature flags but avoid defaults for v2 feature flags.
        if (
            FeatureFlagConstants.ALLOCATION in feature_flag_dict or
            FeatureFlagConstants.VARIANTS in feature_flag_dict or
            FeatureFlagConstants.TELEMETRY in feature_flag_dict
        ):
            feature_flag_version = FeatureFlagVersion.V2

        feature_name = feature_flag_dict.get(FeatureFlagConstants.ID, "")
        if feature_flag_version == FeatureFlagVersion.V1:
            feature_flag_dict[FeatureFlagConstants.DESCRIPTION] = feature_flag_dict.get(FeatureFlagConstants.DESCRIPTION, "")  # assign empty string as default

        if not feature_name:
            raise ValueError("Feature flag 'id' cannot be empty.")

        conditions = feature_flag_dict.get(FeatureFlagConstants.CONDITIONS, None)
        default_conditions = {FeatureFlagConstants.CLIENT_FILTERS: []}

        if conditions:
            client_filters = conditions.get(FeatureFlagConstants.CLIENT_FILTERS, None)

            # Convert all filters to FeatureFilter objects
            if client_filters:
                client_filters_list = []
                for client_filter in client_filters:
                    # If there is a filter, it should always have a name
                    # In case it doesn't, ignore this filter
                    lowercase_filter = {k.lower(): v for k, v in client_filter.items()}
                    name = lowercase_filter.get(FeatureFlagConstants.NAME, None)
                    if name:
                        params = lowercase_filter.get(
                            FeatureFlagConstants.FILTER_PARAMETERS, {}
                        )
                        client_filters_list.append(FeatureFilter(name, params))
                    else:
                        logger.warning(
                            "Ignoring this filter without the %s attribute:\n%s",
                            FeatureFlagConstants.FILTER_NAME,
                            json.dumps(client_filter, indent=2, ensure_ascii=False)
                        )

                conditions[FeatureFlagConstants.CLIENT_FILTERS] = client_filters_list

            requirement_type = conditions.get(
                FeatureFlagConstants.REQUIREMENT_TYPE, None
            )
            if requirement_type:
                # Requirement type is case insensitive.
                if requirement_type.lower() == FeatureFlagConstants.REQUIREMENT_TYPE_ALL.lower():
                    conditions[FeatureFlagConstants.REQUIREMENT_TYPE] = FeatureFlagConstants.REQUIREMENT_TYPE_ALL
                elif requirement_type.lower() == FeatureFlagConstants.REQUIREMENT_TYPE_ANY.lower():
                    conditions[FeatureFlagConstants.REQUIREMENT_TYPE] = FeatureFlagConstants.REQUIREMENT_TYPE_ANY
                else:
                    raise ValidationError(
                        f"Feature '{feature_name}' must have an Any/All requirement type."
                    )

            # Backend returns conditions: {client_filters: None} for flags with no conditions.
            # No need to write empty conditions to key-values.
            if client_filters or requirement_type:
                feature_flag_dict[FeatureFlagConstants.CONDITIONS] = conditions
            else:
                feature_flag_dict[FeatureFlagConstants.CONDITIONS] = default_conditions
        elif feature_flag_version == FeatureFlagVersion.V1:
            feature_flag_dict[FeatureFlagConstants.CONDITIONS] = default_conditions

        # Allocation
        allocation = feature_flag_dict.get(FeatureFlagConstants.ALLOCATION, None)
        if allocation:
            feature_flag_dict[FeatureFlagConstants.ALLOCATION] = FeatureAllocation.convert_from_dict(allocation)

        # Variants
        variants = feature_flag_dict.get(FeatureFlagConstants.VARIANTS, None)
        if variants:
            variant_list = []
            for variant in variants:
                if variant:
                    feature_variant = FeatureVariant.convert_from_dict(variant)
                    variant_list.append(feature_variant)

            feature_flag_dict[FeatureFlagConstants.VARIANTS] = variant_list

        # Telemetry
        telemetry = feature_flag_dict.get(FeatureFlagConstants.TELEMETRY, None)
        if telemetry:
            feature_flag_dict[FeatureFlagConstants.TELEMETRY] = FeatureTelemetry.convert_from_dict(telemetry, feature_name)

        feature_flag_value = FeatureFlagValue(
            id_=feature_name,
            description=feature_flag_dict.get(FeatureFlagConstants.DESCRIPTION, None),
            enabled=feature_flag_dict.get(FeatureFlagConstants.ENABLED, False),
            display_name=feature_flag_dict.get(FeatureFlagConstants.DISPLAY_NAME, None),
            conditions=feature_flag_dict.get(FeatureFlagConstants.CONDITIONS, None),
            allocation=feature_flag_dict.get(FeatureFlagConstants.ALLOCATION, None),
            variants=feature_flag_dict.get(FeatureFlagConstants.VARIANTS, None),
            telemetry=feature_flag_dict.get(FeatureFlagConstants.TELEMETRY, None),
        )

    except (InvalidArgumentValueError, TypeError, ValueError) as exception:
        error_msg = (
            "Invalid value. Unable to decode the following JSON value: \n" +
            "key:{0} value:{1}\nFull exception: \n{2}".format(
                keyvalue.key, keyvalue.value, str(exception)
            )
        )
        raise ValueError(error_msg)

    except:
        logger.error(
            "Exception while parsing feature flag. key:%s value:%s.",
            keyvalue.key,
            keyvalue.value,
        )
        raise

    return feature_flag_value


def is_feature_flag(kv):
    # pylint: disable=line-too-long
    """
    Helper function used to determine if a key-value is a feature flag
    """
    if (
        kv and
        kv.key and
        isinstance(kv.key, str) and
        kv.content_type and
        isinstance(kv.content_type, str)
    ):
        return (
            kv.key.startswith(FeatureFlagConstants.FEATURE_FLAG_PREFIX) and
            kv.content_type == FeatureFlagConstants.FEATURE_FLAG_CONTENT_TYPE
        )
    return False


def convert_string_to_bool(value, feature_name):
    """
    Convert the value to a boolean if it is a string.

    :param Union[str, bool] enabled: Value to be converted.
    :return: Converted value.
    :rtype: bool
    """
    if isinstance(value, bool):
        return value
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    raise ValueError(
        f"Invalid setting 'enabled' with value '{value}' for feature '{feature_name}'."
    )


class FeatureManagementReservedKeywords:
    """
    Feature management keywords used in files in different naming conventions.
    """

    FeatureFlagKeywords = namedtuple(
        "FeatureFlagKeywords",
        [
            "feature_management",
            "enabled_for",
            "requirement_type",
        ],
    )

    PASCAL = FeatureFlagKeywords(
        feature_management="FeatureManagement",
        enabled_for="EnabledFor",
        requirement_type="RequirementType",
    )
    CAMEL = FeatureFlagKeywords(
        feature_management="featureManagement",
        enabled_for="enabledFor",
        requirement_type="requirementType",
    )
    UNDERSCORE = FeatureFlagKeywords(
        feature_management="feature_management",
        enabled_for="enabled_for",
        requirement_type="requirement_type",
    )
    HYPHEN = FeatureFlagKeywords(
        feature_management="feature-management",
        enabled_for="enabled-for",
        requirement_type="requirement-type",
    )

    ALL = (PASCAL, CAMEL, UNDERSCORE, HYPHEN)

    @classmethod
    def get_keywords(cls, naming_convention="pascal"):
        return getattr(cls, naming_convention.upper(), cls.PASCAL)
